import asyncio
import logging
from asyncio import Future
from typing import Awaitable, Callable

import websockets
from pyee import EventEmitter
from websockets.legacy.client import connect

from mokr.connection.base import RemoteConnection
from mokr.connection.devtools import DevtoolsConnection
from mokr.constants import TARGET_ATTACH
from mokr.exceptions import NetworkError


LOGGER = logging.getLogger(__name__)


class Connection(EventEmitter, RemoteConnection):
    def __init__(
        self,
        url: str,
        loop: asyncio.AbstractEventLoop,
        delay: int = 0,
    ) -> None:
        """
        Create remote connection.

        Args:
            url (str): Websocket URL for remote connection.
            loop (asyncio.AbstractEventLoop): Running asyncio loop.
            delay (int, optional): Time in milliseconds to wait before
                handling messages. Defaults to 0.
        """
        super().__init__()
        self._url = url
        self._last_id = 0
        self._callbacks: dict[int, asyncio.Future] = dict()
        self._delay = delay / 1000
        self._loop = loop
        self._sessions: dict[str, DevtoolsConnection] = dict()
        self._connected = False
        self._ws = connect(
            self._url,
            max_size=None,
            loop=self._loop,
            ping_interval=None,
            ping_timeout=None,
        )
        self._recv_fut = self._loop.create_task(self._recv_loop())
        self._close_callback: Callable | None = None
        self.connection: DevtoolsConnection = None

    @property
    def url(self) -> str:
        """Get remote websocket URL."""
        return self._url

    async def _recv_loop(self) -> None:
        async with self._ws as connection:
            self._connected = True
            self.connection = connection
            while self._connected:
                try:
                    resp = await self.connection.recv()
                    if resp:
                        await self._handle_response(resp)
                except (websockets.ConnectionClosed, ConnectionResetError):
                    LOGGER.info('Connection closed.')
                    break
                await asyncio.sleep(0)
        if self._connected:
            self._loop.create_task(self.dispose())

    async def _handle_response(self, response: str) -> None:
        await asyncio.sleep(self._delay)
        self._on_message(response)

    async def _async_send(self, msg: str, callback_id: int) -> None:
        while not self._connected:
            await asyncio.sleep(self._delay)
        try:
            await self.connection.send(msg)
        except websockets.ConnectionClosed:
            LOGGER.warning('Connection closed unexpectedly.')
            callback = self._callbacks.get(callback_id, None)
            if callback and not callback.done():
                callback.set_result(None)
                await self.dispose()

    def _on_successful_response(self, callback: Future, msg: dict) -> None:
        callback.set_result(msg.get('result'))

    def _on_query(self, msg: dict) -> None:
        result, method, params = self._handle_detached_or_received(msg)
        if not result:
            self.emit(method, params)

    async def _on_close(self) -> None:
        if self._close_callback:
            self._close_callback()
            self._close_callback = None
        for callback in self._callbacks.values():
            callback.set_exception(
                self._rewrite_exception(
                    callback.error,
                    f'Protocol error {callback.method}: Target closed.',
                )
            )
        self._callbacks.clear()
        for session in self._sessions.values():
            session._on_closed()
        self._sessions.clear()
        if hasattr(self, 'connection'):
            await self.connection.close()
        if not self._recv_fut.done():
            self._recv_fut.cancel()

    def _set_closed_callback(self, callback: Callable) -> None:
        self._close_callback = callback

    def send(self, method: str, params: dict = None) -> Awaitable[dict]:
        """
        Send message to remote connection via websocket.

        Args:
            method (str): Method to run.
            params (dict, optional): Arguments for method, if any.
                Defaults to None.

        Raises:
            ConnectionError: Raised if the connection is closed.

        Returns:
            Awaitable[dict]: Remote response as dictionary.
        """
        # Detect connection availability from the second transmission.
        if self._last_id and not self._connected:
            raise ConnectionError('Connection is closed.')
        msg = self._prepare_message(method, params)
        self._loop.create_task(self._async_send(msg, self._last_id))
        callback = self._loop.create_future()
        self._callbacks[self._last_id] = callback
        callback.error = NetworkError()
        callback.method = method
        return callback

    async def dispose(self) -> None:
        """Sever all connections."""
        self._connected = False
        await self._on_close()

    async def create_session(self, target_info: dict) -> DevtoolsConnection:
        """
        Create a new `mokr.connection.DevtoolsConnection`.

        Args:
            target_info (dict): Target info from triggered event.

        Returns:
            DevtoolsConnection: New `DevtoolsConnection`.
        """
        resp = await self.send(
            TARGET_ATTACH,
            {'targetId': target_info['targetId']}
        )
        session_id = resp.get('sessionId')
        session = DevtoolsConnection(
            self,
            target_info['type'],
            session_id,
            self._loop,
        )
        self._sessions[session_id] = session
        return session



from __future__ import annotations

import logging
from collections import OrderedDict
from typing import TYPE_CHECKING

from pyee import EventEmitter

from mokr.connection import DevtoolsConnection
from mokr.constants import (
    FRAME_ATTACHED,
    FRAME_DETACHED,
    FRAME_NAVIGATED,
    FRAME_NAVIGATED_IN_DOC,
    LIFECYCLE_EVENT,
    PAGE_FRAME_ATTACHED,
    PAGE_FRAME_DETACHED,
    PAGE_FRAME_NAVIGATED,
    PAGE_FRAME_NAVIGATED_IN_DOC,
    PAGE_FRAME_STOPPED_LOADING,
    PAGE_LIFECYCLE_EVENT,
    RUNTIME_EXECUTION_CONTEXT_CREATED,
    RUNTIME_EXECUTION_CONTEXT_DESTROYED,
    RUNTIME_EXECUTION_CONTEXTS_CLEARED,

)
from mokr.exceptions import ElementHandleError, PageError
from mokr.execution.context import ExecutionContext, JavascriptHandle
from mokr.execution.handle.element import ElementHandle
from mokr.frame.frame import Frame

if TYPE_CHECKING:
    from mokr.browser.page import Page


LOGGER = logging.getLogger(__name__)


class FrameManager(EventEmitter):
    def __init__(
        self,
        client: DevtoolsConnection,
        frame_tree: dict,
        page: Page,
    ) -> None:
        """
        This class is an assistant class that helps with `mokr.frame.Frame`
        management by creating them and by managing
        `mokr.execution.ExecutionContext`s.

        Args:
            client (DevtoolsConnection): `mokr.execution.ExecutionContext` of
                the `mokr.browser.Page` that spawned this.
            frame_tree (dict): A representation of the hierarchy of remote
                frames on a page.
            page (Page): The spawning `mokr.browser.Page` object.
        """
        super().__init__()
        self._client = client
        self._page = page
        self._frames: OrderedDict[str, Frame] = OrderedDict()
        self._main_frame: Frame | None = None
        self._context_id_to_context: dict[str, ExecutionContext] = dict()
        events_to_methods = {
            PAGE_FRAME_ATTACHED: lambda event: self._on_frame_attached(
                event.get('frameId', ''),
                event.get('parentFrameId', ''),
            ),
            PAGE_FRAME_NAVIGATED: lambda event: self._on_frame_navigated(
                event.get('frame'),
            ),
            PAGE_FRAME_NAVIGATED_IN_DOC: (
                lambda event: self._on_frame_navigated_within_document(
                    event.get('frameId'),
                    event.get('url'),
                )
            ),
            PAGE_FRAME_DETACHED: lambda event: self._on_frame_detached(
                event.get('frameId'),
            ),
            PAGE_FRAME_STOPPED_LOADING: (
                lambda event: self._on_frame_stopped_loading(
                    event.get('frameId'),
                )
            ),
            RUNTIME_EXECUTION_CONTEXT_CREATED: (
                lambda event: self._on_execution_context_created(
                    event.get('context'),
                )
            ),
            RUNTIME_EXECUTION_CONTEXT_DESTROYED: (
                lambda event: self._on_execution_context_destroyed(
                    event.get('executionContextId'),
                )
            ),
            RUNTIME_EXECUTION_CONTEXTS_CLEARED: (
                lambda _: self._on_execution_contexts_cleared()
            ),
            PAGE_LIFECYCLE_EVENT: lambda event: self._on_lifecycle_event(event),
        }
        for event, method in events_to_methods.items():
            client.on(event, method)
        self._handle_frame_tree(frame_tree)

    @property
    def main_frame(self) -> Frame | None:
        """The main remote frame (top of the frame tree)."""
        return self._main_frame

    def _on_lifecycle_event(self, event: dict) -> None:
        frame = self._frames.get(event['frameId'])
        if not frame:
            return
        frame._on_lifecycle_event(event['loaderId'], event['name'])
        self.emit(LIFECYCLE_EVENT, frame)

    def _on_frame_stopped_loading(self, frame_id: str) -> None:
        frame = self._frames.get(frame_id)
        if not frame:
            return
        frame._on_loading_stopped()
        self.emit(LIFECYCLE_EVENT, frame)

    def _handle_frame_tree(self, frame_tree: dict) -> None:
        frame = frame_tree['frame']
        if 'parentId' in frame:
            self._on_frame_attached(
                frame['id'],
                frame['parentId'],
            )
        self._on_frame_navigated(frame)
        if 'childFrames' not in frame_tree:
            return
        for child in frame_tree['childFrames']:
            self._handle_frame_tree(child)

    def _on_frame_attached(self, frame_id: str, parent_frame_id: str) -> None:
        if frame_id in self._frames:
            return
        parent_frame = self._frames.get(parent_frame_id)
        frame = Frame(self._client, parent_frame, frame_id)
        self._frames[frame_id] = frame
        self.emit(FRAME_ATTACHED, frame)

    def _on_frame_navigated(self, frame_payload: dict) -> None:
        is_main_frame = not frame_payload.get('parentId')
        if is_main_frame:
            frame = self._main_frame
        else:
            frame = self._frames.get(frame_payload.get('id', ''))
        if not (is_main_frame or frame):
            raise PageError(
                'Either navigated to top level or have a'
                ' stale version of the navigated frame.'
            )
        # Detach all child frames first.
        if frame:
            for child in frame.child_frames:
                self._remove_frames_recursively(child)
        # Update or create main frame.
        _id = frame_payload.get('id', '')
        if is_main_frame:
            if frame:
                # Update id to retain identity on cross-process navigation.
                self._frames.pop(frame._id, None)
                frame._id = _id
            else:
                # Initial main frame navigation.
                frame = Frame(self._client, None, _id)
            self._frames[_id] = frame
            self._main_frame = frame
        # Update frame payload.
        frame._navigated(frame_payload)
        self.emit(FRAME_NAVIGATED, frame)

    def _on_frame_navigated_within_document(
        self,
        frame_id: str,
        url: str,
    ) -> None:
        frame = self._frames.get(frame_id)
        if not frame:
            return
        frame._navigated_within_document(url)
        self.emit(FRAME_NAVIGATED_IN_DOC, frame)
        self.emit(FRAME_NAVIGATED, frame)

    def _on_frame_detached(self, frame_id: str) -> None:
        frame = self._frames.get(frame_id)
        if frame:
            self._remove_frames_recursively(frame)

    def _on_execution_context_created(self, context_payload: dict) -> None:
        aux_data = context_payload.get('auxData')
        if aux_data and aux_data.get('frameId'):
            frame_id = aux_data['frameId']
        else:
            frame_id = None
        frame = self._frames.get(frame_id)
        object_handle_factory = (
            lambda obj: self.create_javascript_handle(
                self.execution_context_by_id(context_payload['id']),
                obj,
            )
        )
        context = ExecutionContext(
            self._client,
            context_payload,
            object_handle_factory,
            frame,
        )
        self._context_id_to_context[context_payload['id']] = context
        if frame:
            frame._add_execution_context(context)

    def _on_execution_context_destroyed(
        self,
        execution_context_id: str,
    ) -> None:
        context = self._context_id_to_context.get(execution_context_id)
        if not context:
            return
        self._context_id_to_context.pop(execution_context_id)
        if context.frame:
            context.frame._remove_execution_context(context)

    def _on_execution_contexts_cleared(self) -> None:
        for context in self._context_id_to_context.values():
            if context.frame:
                context.frame._remove_execution_context(context)
        self._context_id_to_context.clear()

    def _remove_frames_recursively(self, frame: Frame) -> None:
        for child in frame.child_frames:
            self._remove_frames_recursively(child)
        frame._detach()
        self._frames.pop(frame._id, None)
        self.emit(FRAME_DETACHED, frame)

    def frames(self) -> list[Frame]:
        """
        A list of all `mokr.frame.Frame` objects under this manager.

        Returns:
            list[Frame]: All `mokr.frame.Frame`s controlled by this class.
        """
        return list(self._frames.values())

    def frame(self, frame_id: str) -> Frame | None:
        """
        Return a frame with the given `frame_id`, if any.

        Args:
            frame_id (str): Remote frame identifier.

        Returns:
            Frame | None: `mokr.frame.Frame`, if any match given `frame_id`.
        """
        return self._frames.get(frame_id)

    def execution_context_by_id(self, context_id: str) -> ExecutionContext:
        """
        Get a `mokr.execution.ExecutionContext` under this manager.

        Args:
            context_id (str): Target context identifier.

        Raises:
            ElementHandleError: Raised if no context matches `context_id`.

        Returns:
            ExecutionContext: `mokr.execution.ExecutionContext` created
                by this `FrameManager` with given `context_id`.
        """
        context = self._context_id_to_context.get(context_id)
        if not context:
            raise ElementHandleError(f'No context with id of: {context_id}')
        return context

    def create_javascript_handle(
        self,
        context: ExecutionContext,
        remote_object: dict = None,
    ) -> JavascriptHandle:
        """
        Create a `mokr.execution.JavascriptHandle` for given context.

        Args:
            context (ExecutionContext): `mokr.execution.ExecutionContext` to
                pass into initialisation.
            remote_object (dict, optional): Remote object to be represented by
                new `mokr.execution.JavascriptHandle`. Defaults to None.

        Returns:
            JavascriptHandle: _description_
        """
        if remote_object is None:
            remote_object = {}
        if remote_object.get('subtype') == 'node':
            return ElementHandle(
                context,
                self._client,
                remote_object,
                self._page,
                self,
            )
        return JavascriptHandle(context, self._client, remote_object)



import asyncio


class MokrTimeoutError(asyncio.TimeoutError):
    """Timeout Error class."""
    pass


class BaseMokrException(Exception):
    """Base exception for mokr."""
    pass


class ElementHandleError(BaseMokrException):
    """ElementHandle related exception."""
    def __init__(
        self,
        message: str = 'Node is either not visible or not an HTMLElement.',
    ) -> None:
        super().__init__(message)


class NetworkError(BaseMokrException):
    """Network/Protocol related exception."""
    pass


class BrowserError(BaseMokrException):
    """Exception raised from browser."""
    pass


class PageError(BrowserError):
    """Page/Frame related exception."""
    pass


class FirefoxNotImplementedError(NotImplementedError):
    """Exception due to Firefox lacking CDP methods."""
    pass



from __future__ import annotations

import copy
import logging
import math
import os.path
from typing import TYPE_CHECKING, Literal

from mokr.connection import DevtoolsConnection
from mokr.constants import (
    DOM_BOX_MODEL,
    DOM_CONTENT_QUADS,
    DOM_DESCRIBE_NODE,
    DOM_INPUT_FILES,
    METHOD_ELEMENT_IN_VIEW,
    METHOD_EVAL_XPATH,
    METHOD_FOCUS_ELEMENT,
    METHOD_SCROLL_INTO_VIEW,
    PAGE_GET_LAYOUT,
)
from mokr.exceptions import ElementHandleError, NetworkError
from mokr.execution.context import ExecutionContext, JavascriptHandle

if TYPE_CHECKING:
    from mokr.browser.page import Page
    from mokr.frame import Frame, FrameManager


LOGGER = logging.getLogger(__name__)


class ElementHandle(JavascriptHandle):
    def __init__(
        self,
        context: ExecutionContext,
        client: DevtoolsConnection,
        remote_object: dict,
        page: Page,
        frame_manager: FrameManager,
    ) -> None:
        """
        Representation of a DOM element.

        Creation of an `ElementHandle` blocks the remote DOM element from
        garbage collection unless it is disposed manually or the parent frame
        is navigated from.

        Args:
            context (ExecutionContext): `mokr.execution.ExecutionContext` of the
                `mokr.frame.FrameManager` that spawned this.
            client (DevtoolsConnection): Remote connection from parent.
            remote_object (dict): The raw response from the remote connection
                representing the object.
            page (Page): The `mokr.browser.Page` the element was located in.
            frame_manager (FrameManager): The parent `mokr.frame.FrameManager`
                that spawned this element.
        """
        super().__init__(context, client, remote_object)
        self._client = client
        self._remote_object = remote_object
        self._page = page
        self._frame_manager = frame_manager
        self._disposed = False

    @staticmethod
    def _compute_quad_area(quad: list[dict]) -> float:
        area = 0
        for i, _ in enumerate(quad):
            p1 = quad[i]
            p2 = quad[(i + 1) % len(quad)]
            area += (p1['x'] * p2['y'] - p2['x'] * p1['y']) / 2
        return area

    async def _scroll_into_view_if_needed(self) -> None:
        error = await self.execution_context.evaluate(
            METHOD_SCROLL_INTO_VIEW,
            self,
            self._page._javascript_enabled,
        )
        if error:
            raise ElementHandleError(error)

    async def _calculate_origin(self) -> dict[str, float]:
        result = None
        try:
            result = await self._client.send(
                DOM_CONTENT_QUADS,
                {'objectId': self._remote_object.get('objectId')}
            )
        except Exception:
            LOGGER.error(
                "Unhandled error calculating point on element.",
                exc_info=True,
            )
        if not result or not result.get('quads'):
            raise ElementHandleError
        quads = []
        for _quad in result.get('quads'):
            _q = self._from_protocol_quad(_quad)
            if self._compute_quad_area(_q) > 1:
                quads.append(_q)
        if not quads:
            raise ElementHandleError
        quad = quads[0]
        x = 0
        y = 0
        for point in quad:
            x += point['x']
            y += point['y']
        return {'x': x / 4, 'y': y / 4}

    async def _get_box_model(self) -> dict | None:
        try:
            result: dict | None = await self._client.send(
                DOM_BOX_MODEL,
                {'objectId': self._remote_object.get('objectId')},
            )
        except NetworkError:
            LOGGER.error("Error getting box model.", exc_info=True)
            result = None
        return result

    def _from_protocol_quad(self, quad: list[int]) -> list[dict[str, int]]:
        return [
            {'x': quad[0], 'y': quad[1]},
            {'x': quad[2], 'y': quad[3]},
            {'x': quad[4], 'y': quad[5]},
            {'x': quad[6], 'y': quad[7]},
        ]

    def _as_element(self) -> ElementHandle:
        # Used in parity with JavascriptHandle.
        return self

    async def content_frame(self) -> Frame | None:
        """
        Return the content frame for this element or None if not within iframe.

        Returns:
            Frame | None: Target `mokr.frame.Frame` if within iframe.
        """
        node_info = await self._client.send(
            DOM_DESCRIBE_NODE,
            {'objectId': self._remote_object.get('objectId')}
        )
        node_obj = node_info.get('node', {})
        if not isinstance(node_obj.get('frameId'), str):
            return None
        return self._frame_manager.frame(node_obj['frameId'])

    async def content(self) -> str:
        html_handle = await self.get_property("outerHTML")
        value = html_handle._remote_object.get("value")
        return value if value else ""

    async def hover(self) -> None:
        """
        Mouse hover over this element.

        Raises:
            ElementHandleError: Raised if element is detached from DOM.
        """
        await self._scroll_into_view_if_needed()
        center = await self._calculate_origin()
        x = center.get('x', 0)
        y = center.get('y', 0)
        await self._page.mouse.move(x, y)

    async def click(
        self,
        button: Literal["left", "right", "middle"] = "left",
        click_count: int = 1,
        delay: int | float | None = 1000,
    ) -> None:
        """
        Click the center of this element with the bound
        `ElementHandle._page.mouse`.

        If not in view, this will attempt to scroll the element into view.

        Args:
            button (Literal["left", "right", "middle"], optional): Mouse button
                to click with. Defaults to "left".
            click_count (int, optional): Number of clicks to run. Defaults to 1.
            delay (int | float | None, optional): Time in milliseconds to wait
                before each click. Defaults to 1000.

        Raises:
            ElementHandleError: Raised if element is detached from DOM.
        """
        await self._scroll_into_view_if_needed()
        center = await self._calculate_origin()
        x = center.get('x', 0)
        y = center.get('y', 0)
        await self._page.mouse.click(x, y, button, click_count, delay)

    async def upload_file(self, file_paths: list[str]) -> None:
        """
        Set files at given `file_paths` as upload targets for this element.

        Args:
            file_paths (list[str]): List of file paths for upload.
        """
        self._page._is_firefox(caller=self)
        files = [os.path.abspath(p) for p in file_paths]
        object_id = self._remote_object.get('objectId')
        return await self._client.send(
            DOM_INPUT_FILES,
            {'objectId': object_id, 'files': files}
        )

    async def tap(self) -> None:
        """
        Tap the center of this element with the bound
        `ElementHandle._page.touchscreen`.

        If not in view, this will attempt to scroll the element into view.

        Raises:
            ElementHandleError: Raised if element is detached from DOM.
        """
        self._page._is_firefox(caller=self)
        await self._scroll_into_view_if_needed()
        center = await self._calculate_origin()
        x = center.get('x', 0)
        y = center.get('y', 0)
        await self._page.touchscreen.tap(x, y)

    async def focus(self) -> None:
        """Focus on this element."""
        await self.execution_context.evaluate(METHOD_FOCUS_ELEMENT, self)

    async def type_text(self, text: str, delay: int | float = 0) -> None:
        """
        Focus this element and type characters into it. Uses bound
        `ElementHandle._page.keyboard`.

        Note that modifier keys do not alter text case, meaning sending
        `mokr.input.Keyboard.press("shift")` and typing
        `ElementHandle.type_text("mokr")` will not type "MOKR" into the it.

        Args:
            text (str): Text to type.
            delay (int | float, optional): Time in milliseconds to wait between
                each character typed. Defaults to 0.
        """
        await self.focus()
        await self._page.keyboard.type_text(text, delay)

    async def press(self, key: str, delay: int | float | None = None) -> None:
        """
        Focus this element and press the given `key` by sending
        `mokr.input.Keyboard.down` and `mokr.input.Keyboard.up` events, with
        the given `delay` in between.

        Modifier keys effect this method, meaning sending
        `mokr.input.Keyboard.press("shift")` and then
        `mokr.input.Keyboard.press("m")` will type an uppercase "M".

        Args:
            key (str): Name of key. See `mokr.input.Keyboard.key_definitions`.
            delay (int | float | None, optional): Time in milliseconds to "hold"
                the key down. Defaults to None.
        """
        await self.focus()
        await self._page.keyboard.press(key, delay)

    async def bounding_box(self) -> dict[str, float] | None:
        """
        Return the bounding box for this element, if visible. Othwerise,
        return None.

        Returns:
            dict[str, float] | None: Dictionary keyed with "x", "y", "width",
                and "height", if visible. Othwerwise, None.
        """
        result = await self._get_box_model()
        if not result:
            return None
        quad = result['model']['border']
        x = min(quad[0], quad[2], quad[4], quad[6])
        y = min(quad[1], quad[3], quad[5], quad[7])
        width = max(quad[0], quad[2], quad[4], quad[6]) - x
        height = max(quad[1], quad[3], quad[5], quad[7]) - y
        return {'x': x, 'y': y, 'width': width, 'height': height}

    async def box_model(self) -> dict | None:
        """
        Return the box model for this element, if visible. Otherwise, None.

        Returns:
            dict | None: If not visible, None. Othwerwise, Dictionary keyed with
                "content", "padding", "border", "margin", "width", and "height".
                Each key value will be a dictionary containing "x" and "y".
        """
        result = await self._get_box_model()
        if not result:
            return None
        model = result.get('model', {})
        return {
            'content': self._from_protocol_quad(model.get('content')),
            'padding': self._from_protocol_quad(model.get('padding')),
            'border': self._from_protocol_quad(model.get('border')),
            'margin': self._from_protocol_quad(model.get('margin')),
            'width': model.get('width'),
            'height': model.get('height'),
        }

    async def screenshot(
        self,
        file_type: Literal["png", "jpeg"] | None = None,
        file_path: str | None = None,
        jpeg_quality: int | None = None,
        omit_background: bool = False,
        encoding: Literal["binary", "base64"] = "binary",
        scale: int | float = 1,
    ) -> bytes:
        """
        Take a screenshot of the element. Will scroll element inyo the viewport,
        if needed.

        Args:
            file_type (Literal["png", "jpeg"] | None, optional): File type
                to save the image as. Defaults to "png" if None.
            file_path (str | None, optional): Path to save image to. If given
                without `file_type`,  the type will be inferred.
                Defaults to None.
            jpeg_quality (int | None, optional): JPEG quality, 0-100 (only
                applicable if type is or is inferred to be JPEG.
                Defaults to 100 if None.
            omit_background (bool, optional): Make the background transparent
                if True.
                Not supported on Firefox.
                Defaults to False.
            encoding (Literal["binary", "base64"], optional): Encoding type
                to return the image data as. Defaults to "binary".
            scale (int | float, optional): Image scale, 0-1. Defaults to 1.

        Raises:
            ValueError: Raised if `file_type` isn't given and can't be inferred
                as a supported type from `file_path`, if given.
            ElementHandleError: Raised if element is not visible (has no
                `mokr.execution.ElementHandle.bounding_box`).

        Returns:
            bytes: The image content.
        """
        needs_viewport_reset = False
        bounding_box = await self.bounding_box()
        if not bounding_box:
            raise ElementHandleError
        original_viewport = copy.deepcopy(self._page.viewport)
        if (
            bounding_box['width'] > original_viewport['width']
            or bounding_box['height'] > original_viewport['height']
        ):
            new_viewport = {
                'width': max(
                    original_viewport['width'],
                    math.ceil(bounding_box['width']),
                ),
                'height': max(
                    original_viewport['height'],
                    math.ceil(bounding_box['height']),
                ),
            }
            new_viewport = copy.deepcopy(original_viewport)
            new_viewport.update(new_viewport)
            await self._page.set_viewport(new_viewport)
            needs_viewport_reset = True
        await self._scroll_into_view_if_needed()
        bounding_box = await self.bounding_box()
        if not bounding_box:
            raise ElementHandleError
        _obj = await self._client.send(PAGE_GET_LAYOUT)
        page_x = _obj['layoutViewport']['pageX']
        page_y = _obj['layoutViewport']['pageY']
        clip = copy.deepcopy(bounding_box)
        clip['x'] = clip['x'] + page_x
        clip['y'] = clip['y'] + page_y
        buffer = await self._page.screenshot(
            file_type=file_type,
            file_path=file_path,
            jpeg_quality=jpeg_quality,
            clip=clip,
            omit_background=omit_background,
            encoding=encoding,
            scale=scale,
        )
        if needs_viewport_reset:
            await self._page.set_viewport(original_viewport)
        return buffer

    async def query_selector(self, selector: str) -> ElementHandle | None:
        """
        Return the first child element in this element that matches the
        selector, if any.

        Args:
            selector (str): Element selector to locate.

        Returns:
            ElementHandle | None: ElementHandle if found or None.
        """
        handle = await self.execution_context.evaluate_handle(
            '(element, selector) => element.querySelector(selector)',
            self,
            selector,
        )
        element = handle._as_element()
        if element:
            return element
        await handle.dispose()
        return None

    async def query_selector_all(self, selector: str) -> list[ElementHandle]:
        """
        Return all child elements in this element that match the selector,
        if any.

        Args:
            selector (str): Element selector to locate.

        Returns:
            list[ElementHandle]: List of ElementHandle if any or empty list.
        """
        array_handle = await self.execution_context.evaluate_handle(
            '(element, selector) => element.querySelectorAll(selector)',
            self, selector,
        )
        properties = await array_handle.get_properties()
        await array_handle.dispose()
        return [prop._as_element() for prop in properties.values()]

    async def xpath(self, expression: str) -> list[ElementHandle]:
        """
        Return all child elements in this element that match the expression,
        if any.

        Args:
            expression (str): XPath expression to evaluate.

        Returns:
            list[ElementHandle]: List of ElementHandle if any or empty list.
        """
        array_handle = await self.execution_context.evaluate_handle(
            METHOD_EVAL_XPATH,
            self,
            expression,
        )
        properties = await array_handle.get_properties()
        await array_handle.dispose()
        return [prop._as_element() for prop in properties.values()]

    async def is_in_viewport(self) -> bool:
        """
        Evaluate if the element is withing the viewport.

        Returns:
            bool: True if in viewport, else False.
        """
        return await self.execution_context.evaluate(
            METHOD_ELEMENT_IN_VIEW,
            self,
        )



from __future__ import annotations

import asyncio
import gc
import json
import os
import shutil
import socket
import sys
import tempfile
import time
from pathlib import Path
from urllib.parse import urlparse

from geckordp.actors.addon.addons import AddonsActor
from geckordp.actors.preference import PreferenceActor
from geckordp.actors.root import RootActor
from geckordp.rdp_client import RDPClient

from mokr import network
from mokr.browser.browser import Browser
from mokr.constants import INSTALL_PATH
from mokr.exceptions import BrowserError
from mokr.launch.base import Launcher


CHROME_PROFILE_PATH = INSTALL_PATH / '.dev_profile'

# Some of these comments
DEFAULT_FIREFOX_USER_PREFS = {
    # Make sure Shield doesn't hit the network.
    'app.normandy.api_url': '',
    # Disable Firefox old build background check
    'app.update.checkInstallTime': False,
    # Disable automatically upgrading Firefox
    'app.update.disabledForTesting': True,
    # Increase the APZ content response timeout to 1 minute
    'apz.content_response_timeout': 60000,
    # Prevent various error message on the console
    # jest-puppeteer asserts that no error message is emitted by the console
    'browser.contentblocking.features.standard':
    '-tp,tpPrivate,cookieBehavior0,-cm,-fp',
    # Enable the dump function: which sends messages to the system
    # console
    # https://bugzilla.mozilla.org/show_bug.cgi?id=1543115
    'browser.dom.window.dump.enabled': True,
    # Disable topstories
    'browser.newtabpage.activity-stream.feeds.system.topstories': False,
    # Always display a blank page
    'browser.newtabpage.enabled': False,
    # Background thumbnails in particular cause grief: and disabling
    # thumbnails in general cannot hurt
    'browser.pagethumbnails.capturing_disabled': True,
    # Disable safebrowsing components.
    'browser.safebrowsing.blockedURIs.enabled': False,
    'browser.safebrowsing.downloads.enabled': False,
    'browser.safebrowsing.malware.enabled': False,
    'browser.safebrowsing.phishing.enabled': False,
    # Disable updates to search engines.
    'browser.search.update': False,
    # Do not restore the last open set of tabs if the browser has crashed
    'browser.sessionstore.resume_from_crash': False,
    # Skip check for default browser on startup
    'browser.shell.checkDefaultBrowser': False,
    # Disable newtabpage
    'browser.startup.homepage': 'about:blank',
    # Do not redirect user when a milstone upgrade of Firefox is detected
    'browser.startup.homepage_override.mstone': 'ignore',
    # Start with a blank page about:blank
    'browser.startup.page': 0,
    # Do not allow background tabs to be zombified on Android: otherwise for
    # tests that open additional tabs: the test harness tab itself might get
    # unloaded
    'browser.tabs.disableBackgroundZombification': False,
    # Do not warn when closing all other open tabs
    'browser.tabs.warnOnCloseOtherTabs': False,
    # Do not warn when multiple tabs will be opened
    'browser.tabs.warnOnOpen': False,
    # Do not automatically offer translations, as tests do not expect this.
    'browser.translations.automaticallyPopup': False,
    # Disable the UI tour.
    'browser.uitour.enabled': False,
    # Turn off search suggestions in the location bar so as not to trigger
    # network connections.
    'browser.urlbar.suggest.searches': False,
    # Disable first run splash page on Windows 10
    'browser.usedOnWindows10.introURL': '',
    # Do not warn on quitting Firefox
    'browser.warnOnQuit': False,
    # Defensively disable data reporting systems
    'datareporting.healthreport.documentServerURI': "http://dummy.test/dummy/healthreport/",  # noqa
    'datareporting.healthreport.logging.consoleEnabled': False,
    'datareporting.healthreport.service.enabled': False,
    'datareporting.healthreport.service.firstRun': False,
    'datareporting.healthreport.uploadEnabled': False,
    # Do not show datareporting policy notifications
    'datareporting.policy.dataSubmissionEnabled': False,
    'datareporting.policy.dataSubmissionPolicyBypassNotification': True,
    # DevTools JSONViewer sometimes fails to load dependencies with require.js.
    # This doesn't affect Puppeteer but spams console (Bug 1424372)
    'devtools.jsonview.enabled': False,
    # Disable popup-blocker
    'dom.disable_open_during_load': False,
    # Enable the support for File object creation in the content process
    # Required for |Page.setFileInputFiles| protocol method.
    'dom.file.createInChild': True,
    # Disable the ProcessHangMonitor
    'dom.ipc.reportProcessHangs': False,
    # Disable slow script dialogues
    'dom.max_chrome_script_run_time': 0,
    'dom.max_script_run_time': 0,
    # Only load extensions from the application and user profile
    # AddonManager.SCOPE_PROFILE + AddonManager.SCOPE_APPLICATION
    'extensions.autoDisableScopes': 0,
    'extensions.enabledScopes': 5,
    # Disable metadata caching for installed add-ons by default
    'extensions.getAddons.cache.enabled': False,
    # Disable installing any distribution extensions or add-ons.
    'extensions.installDistroAddons': False,
    # Disabled screenshots extension
    'extensions.screenshots.disabled': True,
    # Turn off extension updates so they do not bother tests
    'extensions.update.enabled': False,
    # Turn off extension updates so they do not bother tests
    'extensions.update.notifyUser': False,
    # Make sure opening about:addons will not hit the network
    'extensions.webservice.discoverURL': "http://dummy.test/dummy/discoveryURL",
    # Allow the application to have focus even it runs in the background
    'focusmanager.testmode': True,
    # Disable useragent updates
    'general.useragent.updates.enabled': False,
    # Always use network provider for geolocation tests so we bypass the
    # macOS dialog raised by the corelocation provider
    'geo.provider.testing': True,
    # Do not scan Wifi
    'geo.wifi.scan': False,
    # No hang monitor
    'hangmonitor.timeout': 0,
    # Show chrome errors and warnings in the error console
    'javascript.options.showInConsole': True,
    # Disable download and usage of OpenH264: and Widevine plugins
    'media.gmp-manager.updateEnabled': False,
    # Disable the GFX sanity window
    'media.sanity-test.disabled': True,
    # Disable experimental feature that is only available in Nightly
    'network.cookie.sameSite.laxByDefault': False,
    # Do not prompt for temporary redirects
    'network.http.prompt-temp-redirect': False,
    # Disable speculative connections so they are not reported as leaking
    # when they are hanging around
    'network.http.speculative-parallel-limit': 0,
    # Do not automatically switch between offline and online
    'network.manage-offline-status': False,
    # Make sure SNTP requests do not hit the network
    'network.sntp.pools': 'dummy.test',
    # Disable Flash.
    'plugin.state.flash': 0,
    'privacy.trackingprotection.enabled': False,
    # Can be removed once Firefox 89 is no longer supported
    # https://bugzilla.mozilla.org/show_bug.cgi?id=1710839
    'remote.enabled': True,
    # Don't do network connections for mitm priming
    'security.certerrors.mitm.priming.enabled': False,
    # Local documents have access to all other local documents,
    # including directory listings
    'security.fileuri.strict_origin_policy': False,
    # Do not wait for the notification button security delay
    'security.notification_enable_delay': 0,
    # Ensure blocklist updates do not hit the network
    'services.settings.server': "http://dummy.test/dummy/blocklist/",
    # Do not automatically fill sign-in forms with known usernames and
    # passwords
    'signon.autofillForms': False,
    # Disable password capture, so that tests that include forms are not
    # influenced by the presence of the persistent doorhanger notification
    'signon.rememberSignons': False,
    # Disable first-run welcome page
    'startup.homepage_welcome_url': 'about:blank',
    # Disable first-run welcome page
    'startup.homepage_welcome_url.additional': '',
    # Disable browser animations (tabs, fullscreen, sliding alerts)
    'toolkit.cosmeticAnimations.enabled': False,
    # Prevent starting into safe mode after application crashes
    'toolkit.startup.max_resumed_crashes': -1,
    # Do not close the window when the last tab gets closed
    'browser.tabs.closeWindowWithLastTab': False,
    # Prevent various error message on the console
    # jest-puppeteer asserts that no error message is emitted by the console
    'network.cookie.cookieBehavior': 0,
    # Temporarily force disable BFCache in parent (https://bit.ly/bug-1732263)
    'fission.bfcacheInParent': False,
    # Only enable the CDP protocol
    'remote.active-protocols': 2,
    # Force all web content to use a single content process. TODO: remove
    # this once Firefox supports mouse event dispatch from the main frame
    # context. Once this happens, webContentIsolationStrategy should only
    # be set for CDP. See
    # https://bugzilla.mozilla.org/show_bug.cgi?id=1773393
    'fission.webContentIsolationStrategy': 0,
    # Needed to prevent dialog for connection for loading temporary extensions.
    "devtools.chrome.enabled": True,
    "devtools.debugger.prompt-connection": False,
    # Disable safe-mode on crashes.
    "browser.sessionstore.max_resumed_crashes": 0,
    "browser.sessionstore.restore_on_demand": False,
    "browser.sessionstore.restore_tabs_lazily": False,
    # Enable remote debugging (only needed for loading extensions).
    "devtools.debugger.remote-enabled": True,
    # Set loads of first run / whats new flags.
    "browser.messaging-system.whatsNewPanel.enabled": False,
    "app.normandy.first_run": False,
    "browser.aboutConfig.showWarning": False,
    "browser.aboutwelcome.enabled": False,
    "browser.shell.skipDefaultBrowserCheckOnFirstRun": True,
    "browser.startup.firstrunSkipsHomepage": True,
    "browser.suppress_first_window_animation": True,
    "browser.tabs.warnOnClose": False,
    "devtools.webconsole.timestampMessages": True,
    "doh-rollout.doneFirstRun": True,
    "extensions.formautofill.firstTimeUse": False,
    "pdfjs.firstRun": True,
    "toolkit.telemetry.reportingpolicy.firstRun": False,
    "trailhead.firstrun.branches": "nofirstrun-empty",
    "trailhead.firstrun.didSeeAboutWelcome": True,
}


class FirefoxLauncher(Launcher):
    kind = "firefox"

    def __init__(self, *args, **kwargs) -> None:
        # Flag to indicate that proxy user prefs need to be set via rdp client.
        self._need_set_proxy_via_rdp = False
        super().__init__(*args, **kwargs)
        # async def __aenter__(self, *args, **kwargs) -> FirefoxLauncher:
        self._create_profile()

    def _parse_proxy(self, proxy: str) -> None:
        proxy_parts = urlparse(proxy)
        scheme = proxy_parts.scheme
        host = proxy_parts.hostname
        port = proxy_parts.port
        proxy = f"{scheme}://{host}:{port}"
        credentials = {}
        credentials["proxy"] = proxy
        credentials["host"] = host
        credentials["port"] = port
        if proxy_parts.username:
            credentials["username"] = proxy_parts.username
        if proxy_parts.password:
            credentials["password"] = proxy_parts.password
        return credentials

    def _get_debugger_port(self) -> None:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        for port in range(5999, 65000):
            if port == self.port:
                continue
            try:
                sock = socket.socket()
                sock.bind(('localhost', port))
                sock.close()
                del sock
                gc.collect()
                self._remote_debugger_port = port
                return
            except OSError:
                continue
        raise OSError('No available ports to run debugger.')

    def _make_proxy_prefs(self) -> dict:
        prefs = {}
        proxy_data = self.proxy_credentials
        if proxy_data["proxy"].startswith("socks"):
            prefs.update(
                {
                    "network.proxy.socks": proxy_data["host"],
                    "network.proxy.socks_port": proxy_data["port"],
                    "network.proxy.socks_remote_dns": True,
                }
            )
        else:
            if proxy_data.get("username"):
                # If set here, the debug connection to the RDP session will be
                # blocked by the HTTP proxy.
                # Using this flag to signal to set via RDP after extensions are
                # all ready as RDP will be done with then.
                self._need_set_proxy_via_rdp = True
            else:
                prefs.update(
                    {
                        "network.proxy.http": proxy_data["host"],
                        "network.proxy.ssl": proxy_data["host"],
                        "network.proxy.http_port": proxy_data["port"],
                        "network.proxy.ssl_port": proxy_data["port"],
                    }
                )
        prefs.update(
            {
                "network.proxy.type": 1,
                "network.IDN_show_punycode": True,
                "network.websocket.allowInsecureFromHTTPS": True,
                "network.websocket.auto-follow-http-redirects": True,
            }
        )
        return prefs

    def _create_profile(self):
        # Populates the user.js file with custom preferences as needed to allow
        # Firefox's CDP support to properly function. These preferences will be
        # automatically copied over to prefs.js during startup of Firefox. To be
        # able to restore the original values of preferences a backup of
        # prefs.js will be created.
        extra_prefs = self.firefox_user_prefs if self.firefox_user_prefs else {}
        if not os.path.exists(self.user_data_dir):
            os.mkdir(self.user_data_dir, parents=True)  # recursive: True,
        preferences = DEFAULT_FIREFOX_USER_PREFS.copy()
        preferences.update(extra_prefs)
        if self.proxy_credentials:
            preferences.update(self._make_proxy_prefs())
        lines = [
            f'user_pref({json.dumps(key)}, {json.dumps(value)});'
            for key, value in preferences.items()
        ]
        content = "\n".join(lines)
        with open(os.path.join(self.user_data_dir, "user.js"), "w") as f:
            f.write(content)
        # Create a backup of the preferences file if it already exitsts.
        prefs_path = os.path.join(self.user_data_dir, "prefs.js")
        if os.path.exists(prefs_path):
            prefs_backup_path = os.path.join(
                self.user_data_dir,
                "prefs.js.mokr",
            )
            shutil.copy(prefs_path, prefs_backup_path)

    def _custom_default_args(
        self,
        headless: bool = None,
        user_data_dir: str = None,
        devtools: bool = False,
    ):
        self._get_debugger_port()
        browser_arguments = [
            '-no-remote',
            "-new-instance",
            "-no-default-browser-check",
            # For temporary extensions. Should change to avoid hard-coded port.
            "-start-debugger-server", f"{self._remote_debugger_port}",
        ]
        if headless is None:
            headless = not devtools
        if sys.platform == 'darwin':
            browser_arguments.append('-foreground')
        elif sys.platform == 'win32':
            browser_arguments.append('-wait-for-browser')
        if user_data_dir:
            browser_arguments.append('-profile')
            browser_arguments.append(user_data_dir)
        if headless:
            browser_arguments.append('-headless')
        if devtools:
            browser_arguments.append('-devtools')
        return browser_arguments

    def _compute_launch_args(
        self,
        headless: bool = None,
        user_data_dir: str = None,
        devtools: bool = False,
        ignore_default_args: bool | list[str] = False,
        proxy: str = None,
        args: list[str] = None,
    ):
        super()._compute_launch_args(
            headless=headless,
            user_data_dir=user_data_dir,
            devtools=devtools,
            ignore_default_args=ignore_default_args,
            proxy=proxy,
            args=args,
        )
        self.browser_arguments.append(f"--remote-debugging-port={self.port}")
        self.proxy_credentials = self._parse_proxy(proxy) if proxy else None
        # Check for the profile argument, which will always be set even
        # with a custom directory specified via the user_data_dir option.
        profile_arg_index = next(
            (
                index for index, arg in enumerate(self.browser_arguments)
                if arg in ['-profile', '--profile']
            ),
            None,
        )
        if profile_arg_index is not None:
            data_dir = self.browser_arguments[profile_arg_index + 1]
            if not data_dir or not os.path.exists(data_dir):
                raise ValueError(f"Firefox profile not found at '{data_dir}'")
            # When using a custom Firefox profile it needs to be populated
            # with required preferences.
            self.user_data_dir = data_dir
            self.temp_user_data_dir = None
        else:
            data_dir = tempfile.mkdtemp("mokr_dev_firefox_profile-")
            self.browser_arguments.append('--profile')
            self.browser_arguments.append(data_dir)
            self.temp_user_data_dir = self.user_data_dir = data_dir

    def _clean_restore_data_dirs(self) -> None:
        prefs_backup_path = os.path.join(self.user_data_dir, 'prefs.js.mokr')
        for _ in range(100):
            if self.temp_user_data_dir and os.path.exists(
                self.temp_user_data_dir
            ):
                shutil.rmtree(self.temp_user_data_dir, ignore_errors=True)
                if os.path.exists(self.temp_user_data_dir):
                    time.sleep(0.01)
            elif not self.temp_user_data_dir:
                # When an existing user profile has been used remove the user
                # preferences file and restore possibly backuped preferences.
                if not os.path.exists(prefs_backup_path):
                    return
                try:
                    os.unlink(os.path.join(self.user_data_dir, 'user.js'))
                    if os.path.exists(prefs_backup_path):
                        prefs_path = os.path.join(
                            self.user_data_dir,
                            'prefs.js',
                        )
                        os.unlink(prefs_path)
                        os.rename(prefs_backup_path, prefs_path)
                except Exception:
                    pass
            else:
                return
        else:
            error = (
                'Unable to remove temporary user data dir'
                f' at {self.temp_user_data_dir}'
            )
            if self.temp_user_data_dir:
                raise IOError(error)
            else:
                raise IOError(
                    f"{error} or restore original preferences"
                    f" at {prefs_backup_path}"
                )

    def _make_proxy_extension(self) -> None:
        addon_path = Path(network.__file__).parent / "extensions" / "ffauth"
        # Use an arbitrary name as creds will be within file.
        # TODO load creds to storage so they arent written to disk.
        gen_addon_path = Path(tempfile.mkdtemp())
        shutil.copytree(addon_path, gen_addon_path, dirs_exist_ok=True)
        data = {
            "username": self.proxy_credentials["username"],
            "password": self.proxy_credentials.get("password", ""),
        }
        background_js = gen_addon_path / "background.js"
        content = background_js.read_text()
        new_content = content.replace("creds_placeholder", json.dumps(data))
        background_js.write_text(new_content)
        if not self.firefox_addons_paths:
            self.firefox_addons_paths = []
        self.firefox_addons_paths.insert(0, str(gen_addon_path))

    def _rdp_response_valid(
        self,
        actor_id: str,
        response: dict,
        allow_null: bool = False,
    ) -> bool:
        # https://github.com/jpramosi/geckordp/blob/51e658824d66a2c45b7f4b6e9223703902a1e758/tests/helpers/utils.py#L26  # noqa
        response_string = str(response).lower()
        if not (
            actor_id in response.get("from", "")
            and "no such actor" not in response_string
            and (allow_null or " is null" not in response_string)
            and "unrecognized" not in response_string
        ):
            raise BrowserError("Failed to set Firefox preferences via RDP.")

    def _set_proxy_via_rdp(self, client: RDPClient, root_ids: dict) -> None:
        preference = PreferenceActor(client, root_ids["preferenceActor"])
        prefs = {
            "network.proxy.http": self.proxy_credentials["host"],
            "network.proxy.ssl": self.proxy_credentials["host"],
            "network.proxy.http_port": self.proxy_credentials["port"],
            "network.proxy.ssl_port": self.proxy_credentials["port"],
        }
        for pref, value in prefs.items():
            if isinstance(value, str):
                response = preference.set_char_pref(pref, value)
            elif isinstance(value, int):
                response = preference.set_int_pref(pref, value)
            self._rdp_response_valid("preferenceActor", response)

    def _register_network_addon(self, addon_paths: str) -> None:
        client = RDPClient()
        url_parts = urlparse(self.url)
        host = url_parts.hostname
        port = self._remote_debugger_port
        connection_response = client.connect(host, port)
        if not connection_response:
            raise BrowserError(
                f"Cannot install addons, Firefox unreachable at {host}:{port}."
            )
        root = RootActor(client)
        root_actor_ids = root.get_root()
        addon_actor_id = root_actor_ids["addonsActor"]
        addon_actor = AddonsActor(client, addon_actor_id)
        for addon_path in addon_paths:
            response = addon_actor.install_temporary_addon(addon_path)
            addon_id = response.get("id", None)
            success = addon_id is not None
            if not success:
                raise BrowserError(f"Failed to add extension at {addon_path}.")
        if self._need_set_proxy_via_rdp:
            self._set_proxy_via_rdp(client, root_actor_ids)
        client.disconnect()

    async def launch(self) -> Browser:
        browser = await super().launch()
        if self.proxy_credentials and self.proxy_credentials.get("username"):
            self._make_proxy_extension()
        if self.firefox_addons_paths:
            await asyncio.to_thread(
                self._register_network_addon,
                self.firefox_addons_paths,
            )
        return browser



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from mokr.connection import DevtoolsConnection
from mokr.constants import (
    METHOD_GET_PROPERTY,
    RUNTIME_CALL_FUNCTION,
    RUNTIME_GET_PROPERTIES,
)
from mokr.utils.remote import release_remote_object, serialize_remote_object

if TYPE_CHECKING:
    from mokr.execution.context import ExecutionContext


LOGGER = logging.getLogger(__name__)


class JavascriptHandle():
    def __init__(
        self,
        context: ExecutionContext,
        client: DevtoolsConnection,
        remote_object: dict,
    ) -> None:
        """
        Representative of a JavaScript object within the frame.
        JavascriptHandles are created automatically by `mokr.frame.FrameManager`
        and `mokr.browser.WebWorker`.

        Args:
            context (ExecutionContext): `mokr.execution.ExecutionContext` of the
                `mokr.frame.FrameManager` that spawned this.
            client (DevtoolsConnection): Remote connection from parent.
            remote_object (dict): The raw response from the remote connection
                representing the object.
        """
        self._context = context
        self._client = client
        self._remote_object = remote_object
        self._disposed = False

    @property
    def execution_context(self) -> ExecutionContext:
        """Get `mokr.execution.ExecutionContext` of this handle."""
        return self._context

    def _as_element(self) -> None:
        # Used in parity with ElementHandle.
        return None

    async def get_property(self, property_name: str) -> JavascriptHandle:
        """
        Get a property of the remote object related to this handle.

        Args:
            property_name (str): Name of the target property.

        Returns:
            JavascriptHandle: Newly created `JavascriptHandle` from the result.
        """
        object_handle = await self._context.evaluate_handle(
            METHOD_GET_PROPERTY,
            self,
            property_name,
        )
        properties = await object_handle.get_properties()
        result = properties[property_name]
        await object_handle.dispose()
        return result

    async def get_properties(self) -> dict[str, JavascriptHandle]:
        """
        Get all properties of the remote object related to this handle.

        Returns:
            dict[str, JavascriptHandle]: Dictionary keyed as "property name" to
                newly created `JavascriptHandle`s from the results.
        """
        response = await self._client.send(
            RUNTIME_GET_PROPERTIES,
            {
                'objectId': self._remote_object.get('objectId', ''),
                'ownProperties': True,
            }
        )
        result = {}
        for prop in response['result']:
            if not prop.get('enumerable'):
                continue
            result[prop.get('name')] = self._context._object_handle_factory(
                prop.get('value')
            )
        return result

    async def json(self) -> dict:
        """
        Get and JSONify the values of the remote object this handle represents.

        Returns:
            dict: Dictionary of JSONified values.
        """
        object_id = self._remote_object.get('objectId')
        if object_id:
            response = await self._client.send(
                RUNTIME_CALL_FUNCTION,
                {
                    'functionDeclaration': 'function() { return this; }',
                    'objectId': object_id,
                    'returnByValue': True,
                    'awaitPromise': True,
                }
            )
            return serialize_remote_object(response['result'])
        return serialize_remote_object(self._remote_object)

    async def dispose(self) -> None:
        """
        Stop referencing this handle. Allows it to be garbage-collected by the
        browser.
        """
        if self._disposed:
            return
        self._disposed = True
        try:
            await release_remote_object(
                self._client,
                self._remote_object,
            )
        except Exception:
            LOGGER.error("Error disposing element.", exc_info=True)

    def to_string(self) -> str:
        """
        Get a string representation of the remote object this represents.

        Returns:
            str: String representation of remote object. May contain
                object data or just the type.
        """
        if self._remote_object.get('objectId'):
            _type = (
                self._remote_object.get('subtype')
                or self._remote_object.get('type')
            )
            return f'JavascriptHandle@{_type}'
        return 'JavascriptHandle:{}'.format(
            serialize_remote_object(self._remote_object)
        )



from __future__ import annotations

import asyncio
import logging
from typing import Any, Awaitable, Literal, Set

from mokr.connection import DevtoolsConnection
from mokr.constants import (
    METHOD_EMBED_JAVASCRIPT_BY_CONTENT,
    METHOD_EMBED_JAVASCRIPT_BY_URL,
    METHOD_EMBED_STYLE_BY_CONTENT,
    METHOD_EMBED_STYLE_BY_URL,
    METHOD_GET_CONTENT,
    METHOD_SELECT_VALUES,
    METHOD_SET_CONTENT,
    METHOD_WAIT_FOR_XPATH_OR_SELECTOR,
)
from mokr.exceptions import ElementHandleError, PageError
from mokr.execution.context import ExecutionContext, JavascriptHandle
from mokr.execution.handle.element import ElementHandle
from mokr.frame.wait import WaitTask


LOGGER = logging.getLogger(__name__)


class Frame():
    def __init__(
        self,
        client: DevtoolsConnection,
        parent_frame: Frame | None,
        frame_id: str,
    ) -> None:
        """
        Representative of a frame within a page, spawned from a
        `mokr.frame.FrameManager` in a `mokr.browser.Page`.

        Args:
            client (DevtoolsConnection): `mokr.execution.ExecutionContext` of
                the `mokr.frame.FrameManager` that spawned this.
            parent_frame (Frame | None): Parent `Frame` in tree, if any.
            frame_id (str): Unique remote identifier.
        """
        self._client = client
        self._parent_frame = parent_frame
        self._url = ''
        self._detached = False
        self._id = frame_id
        self._document_promise: ElementHandle | None = None
        self._context_resolve_callback = lambda _: None
        self._navigation_url = None
        self._set_default_context(None)
        self._wait_tasks: Set[WaitTask] = set()
        self._loader_id = ''
        self._lifecycle_events: Set[str] = set()
        self._child_frames: Set[Frame] = set()
        if self._parent_frame:
            self._parent_frame._child_frames.add(self)

    @property
    def name(self) -> str:
        """Get the name of the remote frame."""
        return self.__dict__.get('_name', '')

    @property
    def url(self) -> str:
        """The URL for the remote frame."""
        return self._url

    @property
    def parent_frame(self) -> Frame | None:
        """
        The parent `Frame` in the tree. If the remote frame is the top or is
        detached, will return None.
        """
        return self._parent_frame

    @property
    def child_frames(self) -> list[Frame]:
        """A list of all child `Frame`s in the tree."""
        return list(self._child_frames)

    @property
    def is_detached(self) -> bool:
        """True if remote frame is detached to the page, otherwise False."""
        return self._detached

    def _add_execution_context(self, context: ExecutionContext) -> None:
        if context._is_default:
            self._set_default_context(context)

    def _remove_execution_context(self, context: ExecutionContext) -> None:
        if context._is_default:
            self._set_default_context(None)

    def _set_default_context(self, context: ExecutionContext | None) -> None:
        if context is not None:
            self._context_resolve_callback(context)
            self._context_resolve_callback = lambda _: None
            for wait_task in self._wait_tasks:
                self._client._loop.create_task(wait_task.rerun())
        else:
            self._document_promise = None
            self._context_promise = self._client._loop.create_future()
            self._context_resolve_callback = (
                lambda _context: self._context_promise.set_result(_context)
            )

    def _wait_for_selector_or_xpath(
        self,
        selector_or_xpath: str,
        is_xpath: bool,
        visible: bool = False,
        hidden: bool = False,
        timeout: int = 30000,
    ) -> WaitTask:
        polling = 'raf' if hidden or visible else 'mutation'
        title = (
            f'{"XPath" if is_xpath else "selector"} "{selector_or_xpath}"'
            f'{" to be hidden" if hidden else ""}'
        )
        return WaitTask(
            self,
            METHOD_WAIT_FOR_XPATH_OR_SELECTOR,
            title,
            polling,
            timeout,
            self._client._loop,
            selector_or_xpath,
            is_xpath,
            visible,
            hidden,
        )

    def _navigated(self, frame_payload: dict) -> None:
        self._name = frame_payload.get('name', '')
        self._navigation_url = frame_payload.get('url', '')
        self._url = frame_payload.get('url', '')

    def _navigated_within_document(self, url: str) -> None:
        self._url = url

    def _on_lifecycle_event(self, loader_id: str, name: str) -> None:
        if name == 'init':
            self._loader_id = loader_id
            self._lifecycle_events.clear()
        else:
            self._lifecycle_events.add(name)

    def _on_loading_stopped(self) -> None:
        self._lifecycle_events.add('DOMContentLoaded')
        self._lifecycle_events.add('load')

    def _detach(self) -> None:
        for wait_task in self._wait_tasks:
            wait_task.terminate(
                PageError('Method wait_for_* failed: Frame detached.'))
        self._detached = True
        if self._parent_frame:
            self._parent_frame._child_frames.remove(self)
        self._parent_frame = None

    async def _ensure_execution_context(self) -> ExecutionContext:
        context = await self.execution_context()
        if context is None:
            raise PageError('Frame has no context.')
        return context

    async def _ensure_handle(self, selector: str) -> ElementHandle:
        handle = await self.query_selector(selector)
        if not handle:
            raise PageError(f'No node found for selector: {selector}')
        return handle

    async def _document(self) -> ElementHandle:
        if self._document_promise:
            return self._document_promise
        context = await self._ensure_execution_context()
        document = (await context.evaluate_handle('document'))._as_element()
        self._document_promise = document
        if document is None:
            raise PageError('Could not find document.')
        return document

    async def execution_context(self) -> ExecutionContext | None:
        """
        Return the newly created `mokr.execution.ExecutionContext`.
        Created by `mokr.frame.FrameManager`.

        Returns:
            ExecutionContext: `mokr.execution.ExecutionContext`.
        """
        return await self._context_promise

    async def evaluate_handle(
        self,
        page_function: str,
        *args: Any,
    ) -> JavascriptHandle:
        """
        Execute a JavaScript function with given arguments.
        Runs this `Frame`'s `mokr.execution.ExecutionContext.evaluate_handle`.

        Args:
            page_function (str): JavaScript function to run.

        Returns:
            JavascriptHandle: `mokr.execution.JavascriptHandle`.
        """
        context = await self._ensure_execution_context()
        return await context.evaluate_handle(page_function, *args)

    async def evaluate(
        self,
        page_function: str,
        *args: Any,
        force_expr: bool = False,
    ) -> Any:
        """
        Execute a JavaScript function with given arguments.
        Runs this `Frame`'s `mokr.execution.ExecutionContext.evaluate`.

        Args:
            page_function (str): JavaScript function to run.
            force_expr (bool): If True, treat `page_function` as an expression.
                Otherwise, automatically determine if it is a function or
                an expression. Defaults to False.

        Raises:
            `mokr.exceptions.NetworkError` if an unhandled error occurs
            either evaluating the function or requesting the resulting object.

        Returns:
            dict | None: The decoded object (dict) via
                `mokr.execution.JavascriptHandle.json` or None if a known
                error occurs decoding it.
        """
        context = await self._ensure_execution_context()
        return await context.evaluate(
            page_function,
            *args,
            force_expr=force_expr,
        )

    async def query_selector(self, selector: str) -> ElementHandle | None:
        """
        Return the first element in the DOM that matches the selector, if any.

        Args:
            selector (str): Element selector to locate.

        Returns:
            ElementHandle | None: ElementHandle if found or None.
        """
        document = await self._document()
        value = await document.query_selector(selector)
        return value

    async def query_selector_all(self, selector: str) -> list[ElementHandle]:
        """
        Return all elements in the DOM that match the selector, if any.

        Args:
            selector (str): Element selector to locate.

        Returns:
            list[ElementHandle]: List of ElementHandle if any or empty list.
        """
        document = await self._document()
        value = await document.query_selector_all(selector)
        return value

    async def xpath(self, expression: str) -> list[ElementHandle]:
        """
        Return all elements in the DOM that match the expression, if any.

        Args:
            expression (str): XPath expression to evaluate.

        Returns:
            list[ElementHandle]: List of ElementHandle if any or empty list.
        """
        document = await self._document()
        value = await document.xpath(expression)
        return value

    async def content(self) -> str:
        """
        Get encoded string representation of HTML in this `Frame`.

        Returns:
            str: HTML content.
        """
        return await self.evaluate(METHOD_GET_CONTENT.strip())

    async def set_content(self, html: str) -> None:
        """
        Set the value of the HTML on this `Frame`.

        This does not change the value of the "document" request's response,
        it only changes what is shown in the window.

        Response values can be changed by intercepting requests or responses,
        see `Page.on` for more.

        Args:
            html (str): HTML content to set to.
        """
        await self.evaluate(METHOD_SET_CONTENT, html)

    async def embed_javascript(
        self,
        file_content: str = None,
        file_path: str = None,
        url: str = None,
        script_type: str = None,
    ) -> ElementHandle:
        """
        Add script tag to this `Frame`.

        Args:
            file_content (str, optional): Encoded file content for script to
                embed. If not given, must give `file_path` or `url`. Defaults
                to None.
            file_path (str, optional): File path for script to embed.
                If not given, must give `file_content` or `url`. Defaults to
                None.
            url (str, optional): URL for script to embed. If not given, must
                give `file_content` or `file_path`. Defaults to None.
            script_type (str, optional): Use "module" to load as JavaScript
                ES6 module, if not given, defaults to "text/javascript".
                Defaults to None.

        Raises:
            ValueError: Raised if none of `file_content`, `file_path`, or
                `url` are given.
            PageError: Raised if error occurs sending embed request to remote
                connection (from ElementHandleError).

        Returns:
            ElementHandle: Newly embedded element.
        """
        if not any(url, file_path, file_content):
            raise ValueError("Must provide url, file_path, or file_content.")
        context = await self._ensure_execution_context()
        if file_content:
            args = [METHOD_EMBED_JAVASCRIPT_BY_CONTENT, file_content]
        elif file_path:
            with open(file_path) as f:
                contents = f.read()
            contents = contents + '//# sourceURL={}'.format(
                file_path.replace('\n', '')
            )
            args = [METHOD_EMBED_JAVASCRIPT_BY_CONTENT, contents]
        elif url:
            args = [METHOD_EMBED_JAVASCRIPT_BY_URL, url]
        if script_type:
            args.append(script_type)
        try:
            return (await context.evaluate_handle(*args))._as_element()
        except ElementHandleError:
            raise PageError('Failed to embed script, likely a network error.')

    async def embed_style(
        self,
        file_content: str = None,
        file_path: str = None,
        url: str = None,
    ) -> ElementHandle:
        """
        Add style tag to this `Frame`.

        Args:
            file_content (str, optional): Encoded file content for style to
                embed. If not given, must give `file_path` or `url`. Defaults
                to None.
            file_path (str, optional): File path for style to embed.
                If not given, must give `file_content` or `url`. Defaults to
                None.
            url (str, optional): URL for style to embed. If not given, must
                give `file_content` or `file_path`. Defaults to None.

        Raises:
            ValueError: Raised if none of `file_content`, `file_path`, or
                `url` are given.
            PageError: Raised if error occurs sending embed request to remote
                connection (from ElementHandleError).

        Returns:
            ElementHandle: Newly embedded element.
        """
        if not any(url, file_path, file_content):
            raise ValueError("Must provide url, file_path, or file_content.")
        context = await self._ensure_execution_context()
        if file_content:
            args = [METHOD_EMBED_STYLE_BY_CONTENT, file_content]
        elif file_path:
            with open(file_path) as f:
                contents = f.read()
            contents = contents + '/*# sourceURL={}*/'.format(
                file_path.replace('\n', '')
            )
            args = [METHOD_EMBED_STYLE_BY_CONTENT, contents]
        elif url:
            args = [METHOD_EMBED_STYLE_BY_URL, url]
        try:
            return (await context.evaluate_handle(*args))._as_element()
        except ElementHandleError:
            raise PageError('Failed to embed style, likely a network error.')

    async def click(
        self,
        selector: str,
        button: Literal["left", "right", "middle"] = "left",
        click_count: int = 1,
        delay: int | float | None = 1000,
    ) -> None:
        """
        Click the first element that matches `selector`.

        This method is a shortcut for running `Frame.query_selector` and then
        running `mokr.execution.ElementHandle.click` on the resultant
        ElementHandle. In either case, an element will be scrolled into view
        if needed, and the center of it clicked with the ElementHandle's
        bound `Page.mouse`.

        Args:
            selector (str): Selector to query element by.
            button (Literal["left", "right", "middle"], optional): Mouse button
                to click with. Defaults to "left".
            click_count (int, optional): Number of clicks to run. Defaults to 1.
            delay (int | float | None, optional): Time in milliseconds to wait
                before each click. Defaults to 1000.

        Raises:
            PageError: Raised if no element is found with given `selector`.
        """
        handle = await self._ensure_handle(selector)
        await handle.click(button, click_count, delay)
        await handle.dispose()

    async def focus(self, selector: str) -> None:
        """
        Focus on the first element that matches `selector`.

        Raises:
            PageError: Raised if no element is found with given `selector`.

        Args:
            selector (str): Selector to query element by.
        """
        handle = await self._ensure_handle(selector)
        await handle.focus()
        await handle.dispose()

    async def hover(self, selector: str) -> None:
        """
        Mouse hover over the first element that matches `selector`.

        Raises:
            PageError: Raised if no element is found with given `selector`.

        Args:
            selector (str): Selector to query element by.
        """
        handle = await self._ensure_handle(selector)
        await handle.hover()
        await handle.dispose()

    async def select(self, selector: str, values: list[str]) -> list[str]:
        """
        Select options on a "select" element.

        Args:
            selector (str): Selector to query element by.
            values (list[str]): List of string options to select by.

        Returns:
            list[str]: List of selected values.
        """
        if any(not isinstance(value, str) for value in values):
            raise TypeError("Select values must all be str.")
        handle = await self._ensure_handle(selector)
        context = await self._ensure_execution_context()
        result = context.evaluate(METHOD_SELECT_VALUES, handle, *values)
        await handle.dispose()
        return result

    async def tap(self, selector: str) -> None:
        """
        Tap the first element that matches `selector`.

        This method is a shortcut for running `Frame.query_selector` and then
        running `mokr.execution.ElementHandle.tap` on the resultant
        ElementHandle. In either case, an element will be scrolled into view
        if needed, and the center of it clicked with the ElementHandle's
        bound `Page.touchscreen`.

        Args:
            selector (str): Selector to query element by.

        Raises:
            PageError: Raised if no element is found with given `selector`.
        """
        handle = await self._ensure_handle(selector)
        await handle.tap()
        await handle.dispose()

    async def type_text(
        self,
        selector: str,
        text: str,
        delay: int | float = 0,
    ) -> None:
        """
        Focus on the first element that matches `selector` and type characters
        into it. Uses the newly created `ElementHandle`'s bound `Page.keyboard`.

        Note that modifier keys do not alter text case, meaning sending
        `mokr.input.Keyboard.press("shift")` and typing
        `Frame.type_text("input", "mokr")` will not type "MOKR" into the it.

        Raises:
            PageError: Raised if no element is found with given `selector`.

        Args:
            selector (str): Selector to query element by.
            text (str): Text to type.
            delay (int | float, optional): Time in milliseconds to wait between
                each character typed. Defaults to 0.
        """
        handle = await self._ensure_handle(selector)
        await handle.type_text(text, delay)
        await handle.dispose()

    def wait_for_timeout(
        self,
        timeout: int | float,
    ) -> Awaitable[None]:
        """
        Wait for the given amount of time. Same as `asyncio.sleep`.

        Args:
            timeout (int | float): Time in milliseconds to wait.

        Returns:
            Awaitable[None]: Task to be awaited.
        """
        return self._client._loop.create_task(asyncio.sleep(timeout / 1000))

    def wait_for_selector(
        self,
        selector: str,
        visible: bool = False,
        hidden: bool = False,
        timeout: int = 30000,
    ) -> WaitTask:
        """
        Wait for element that matches `selector` to appear in DOM.
        If element is in DOM already when called, return immediately.

        Args:
            selector (str): Selector to query element by.
            visible (bool, optional): Element must also not be hidden.
                Defaults to False.
            hidden (bool, optional): Element must also be hidden.
                Defaults to False.
            timeout (int, optional): Time in milliseconds to wait.
                Defaults to 30000.

        Raises:
            MokrTimeoutError: Raised if timeout exceeded before element found.
        Returns:
            Awaitable[JavascriptHandle]: None.
        """
        return self._wait_for_selector_or_xpath(
            selector,
            False,
            visible,
            hidden,
            timeout,
        )

    def wait_for_xpath(
        self,
        xpath: str,
        visible: bool = False,
        hidden: bool = False,
        timeout: int = 30000,
    ) -> WaitTask:
        """
        Wait for element that matches `xpath` expression to appear in DOM.
        If element is in DOM already when called, return immediately.

        Args:
            xpath (str): Expression to query element by.
            visible (bool, optional): Element must also not be hidden.
                Defaults to False.
            hidden (bool, optional): Element must also be hidden.
                Defaults to False.
            timeout (int, optional): Time in milliseconds to wait.
                Defaults to 30000.

        Raises:
            MokrTimeoutError: Raised if timeout exceeded before element found.
        Returns:
            Awaitable[JavascriptHandle]: None.
        """
        return self._wait_for_selector_or_xpath(
            xpath,
            True,
            visible,
            hidden,
            timeout,
        )

    def wait_for_function(
        self,
        page_function: str,
        polling: Literal['raf', 'mutation'] | int | float,
        timeout: int = 30000,
    ) -> Awaitable[JavascriptHandle]:
        """
        Wait until the given `page_function` returns a truthy value.

        Args:
            page_function (str): JavaScript function to run.
            polling (Literal["raf", "mutation"] | int | float, optional):
                Polling type; if set to "raf", executes continously in
                "requestAnimationFrame", else if set to "mutation" executes
                only on DOM mutations.
                Defaults to "raf".
            timeout (int, optional): Time in milliseconds to wait.
                Defaults to 30000.

        Returns:
            Awaitable[JavascriptHandle]: JavascriptHandle from JavaScript
                `page_function` successful result.
        """
        return WaitTask(
            self,
            page_function,
            'function',
            polling,
            timeout,
            self._client._loop,
        )

    async def title(self) -> str:
        """
        Get the document title.

        Returns:
            str: Document title.
        """
        return await self.evaluate('() => document.title')



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, Callable, Literal

from mokr.browser.page import Page
from mokr.connection import DevtoolsConnection

if TYPE_CHECKING:
    from mokr.browser import Browser
    from mokr.browser.context import BrowserContext


KINDS = Literal["page", "background_page", "service_worker", "browser", "other"]


class Target():
    def __init__(
        self,
        browser: Browser,
        target_info: dict,
        browser_context: BrowserContext,
        session_factory: Callable,
        ignore_https_errors: bool,
        default_viewport: dict | None,
        screenshot_task_queue: list,
        loop: asyncio.AbstractEventLoop,
        proxy_credentials: dict | None = None,
        user_agent_data: dict[str, str] | None = None,
    ) -> None:
        """
        Representative of a remote target.
        Targets are a somewhat obscure concept, they can be a browser, frame,
        page, worker, or more. For the purposes of this package, only
        browser, pages, and workers are tracked explicitly.
        Full list of target types can be found at `devtools_agent_host_impl.cc`
        at https://source.chromium.org/chromium/chromium/src/.

        Args:
            browser (Browser): Parent `mokr.browser.Browser`.
            target_info (dict): The "targetInfo" in the event emitted that
                triggered this object's initialisation.
            browser_context (BrowserContext): The `mokr.browser.BrowserContext`
                that this spawned from.
            session_factory (Callable): Callable that creates a new
                `mokr.connection.DevtoolsConnection` for this target.
            ignore_https_errors (bool): Whether network security errors should
                be ignored or not. Inherited from parent `mokr.browser.Browser`.
            default_viewport (dict | None): Default viewport configuration.
                Inherited from parent `mokr.browser.Browser`.
            screenshot_task_queue (list): The screenshot task queue, empty by
                default. Inherited from parent `mokr.browser.Browser`.
            loop (asyncio.AbstractEventLoop): The `asyncio` loop that is tracked
                under the `mokr.browser.Browser`'s `mokr.connection.Connection`.
            proxy_credentials: (dict | None, optional): Dictionary with proxy
                credentials keyed as "username" and "password". Credentials
                should be for the proxy the browser process is bound to.
                Inherited from parent `mokr.browser.Browser`.
            user_agent_data (dict[str, str] | None, optional): A dictionary
                containing the user agent and an indicator to whether it was the
                original or an override.
        """
        self._browser = browser
        self._target_info = target_info
        self._browser_context = browser_context
        self._targetId = target_info.get('targetId', '')
        self._session_factory = session_factory
        self._ignore_https_errors = ignore_https_errors
        self._default_viewport = default_viewport
        self._user_agent_data = user_agent_data
        self._screenshot_task_queue = screenshot_task_queue
        self._loop = loop
        self._proxy_credentials = proxy_credentials
        self._page: Page | None = None
        self._initialized_promise = self._loop.create_future()
        self._is_closed_promise = self._loop.create_future()
        self._is_initialized = (
            self._target_info['type'] != 'page'
            or self._target_info['url'] != ''
        )
        if self._is_initialized:
            self._initialized_callback(True)

    @property
    def url(self) -> str:
        """Get url of this target."""
        return self._target_info['url']

    @property
    def kind(self) -> KINDS:
        """
        Get type of this target.
        Type may be on of "page", "background_page", "service_worker",
        "browser", or "other".
        """
        _type = self._target_info['type']
        if _type in ['page', 'background_page', 'service_worker', 'browser']:
            return _type
        return 'other'

    @property
    def browser(self) -> Browser:
        """
        The `mokr.browser.Browser` that owns the `mokr.browser.BrowserContext`
        that this target was initialised from.
        """
        return self._browser_context.browser

    @property
    def browser_context(self) -> BrowserContext:
        """`mokr.browser.BrowserContext` this target was initialised from."""
        return self._browser_context

    @property
    def opener(self) -> Target | None:
        """The parent `Target` that spawned this, if any."""
        opener_id = self._target_info.get('openerId')
        if opener_id is None:
            return None
        return self.browser._targets.get(opener_id)

    def _target_info_changed(self, target_info: dict) -> None:
        self._target_info = target_info
        if not self._is_initialized and (
            self._target_info['type'] != 'page'
            or self._target_info['url'] != ''
        ):
            self._is_initialized = True
            self._initialized_callback(True)
            return

    def _initialized_callback(self, result: bool) -> None:
        if self._initialized_promise.done():
            self._initialized_promise = self._loop.create_future()
        self._initialized_promise.set_result(result)

    def _closed_callback(self) -> None:
        self._is_closed_promise.set_result(None)

    async def create_devtools_connection(self) -> DevtoolsConnection:
        """
        Initialise the `mokr.connection.DevtoolsConnection` that will be bound
        to this `Target`.

        Returns:
            DevtoolsConnection: The new `mokr.connection.DevtoolsConnection`.
        """
        return await self._session_factory()

    async def page(self) -> Page | None:
        """
        Get the `mokr.browser.Page` object attached to this `Target`.
        Only applicable if `Target.kind` is "page" or "background_page".

        Returns:
            Page | None: Associated `mokr.browser.Page`, if any.
        """
        if (
            self._target_info['type'] in ['page', 'background_page']
            and self._page is None
        ):
            client = await self._session_factory()
            new_page = await Page.create(
                self._browser,
                client,
                self,
                self._ignore_https_errors,
                self._default_viewport,
                self._screenshot_task_queue,
                self._proxy_credentials,
                self._user_agent_data,
            )
            self._page = new_page
            return new_page
        return self._page



import json
import logging
from abc import ABC
from typing import Awaitable

from mokr.constants import TARGET_DETACHED, TARGET_RECV_MSG


LOGGER = logging.getLogger(__name__)


class RemoteConnection(ABC):
    @staticmethod
    def _rewrite_exception(error: Exception, message: str) -> Exception:
        error.args = (message, )
        return error

    def _create_protocol_exception(
        self,
        error: Exception,
        method: str,
        obj: dict
    ) -> Exception:
        message = f'Protocol error ({method}): {obj["error"]["message"]}'
        if 'data' in obj['error']:
            message += f' {obj["error"]["data"]}'
        return self._rewrite_exception(error, message)

    def _prepare_message(self, method: str, params: dict = None) -> Awaitable:
        if params is None:
            params = {}
        self._last_id += 1
        msg = json.dumps(
            dict(
                id=self._last_id,
                method=method,
                params=params,
            )
        )
        LOGGER.debug(f'Prepared remote connection message: {msg}')
        return msg

    def _handle_detached_or_received(self, msg: dict) -> tuple[bool, dict, str]:
        # Handle message if received or detached.
        # Returns a tuple of values: result, method, and params.
        # The result is only True when an unhandled method is given.
        method = msg.get('method', '')
        params = msg.get('params', {})
        session_id = params.get('sessionId')
        session = self._sessions.get(session_id)
        if method == TARGET_RECV_MSG and session:
            session._on_message(params.get('message'))
            return True, method, params
        elif method == TARGET_DETACHED and session:
            session._on_closed()
            self._sessions.pop(session_id)
            return True, method, params
        elif method not in (TARGET_DETACHED, TARGET_RECV_MSG):
            return False, method, params
        else:
            return True, method, params

    def _on_message(self, message: str) -> None:
        LOGGER.debug(f'Loading remote connection message: {message}')
        msg = json.loads(message)
        _id = msg.get('id')
        if _id in self._callbacks:
            callback = self._callbacks.pop(_id)
            if msg.get('error'):
                callback.set_exception(
                    self._create_protocol_exception(
                        callback.error,
                        callback.method,
                        msg,
                    )
                )
            else:
                self._on_successful_response(callback, msg)
        else:
            self._on_query(msg)



from typing import Literal


LIFECYCLE_EVENTS = Literal[
    "load",
    "domcontentloaded",
    "networkidle",
    "networkalmostidle",
]

HTTP_METHODS = Literal[
    "GET",
    "POST",
    "OPTIONS",
    "HEAD",
    "PUT",
    "PATCH",
    "DELETE",
]

HTTP_CACHE_TYPES = Literal[
    "default",
    "no-store",
    "reload",
    "no-cache",
    "force-cache",
    "only-if-cache",
]

REFERRER_POLICIES = Literal[
    "no-referrer",
    "no-referrer-when-downgrade",
    "same-origin",
    "origin",
    "strict-origin",
    "origin-when-cross-origin",
    "strict-origin-when-cross-origin",
    "unsafe-url",
]



from mokr.browser.browser import Browser  # noqa
from mokr.browser.console import ConsoleMessage  # noqa
from mokr.browser.context import BrowserContext  # noqa
from mokr.browser.page import Page  # noqa
from mokr.browser.target import Target  # noqa
from mokr.browser.viewport import ViewportManager  # noqa
from mokr.browser.worker import WebWorker  # noqa



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from pyee import EventEmitter

from mokr.browser.page import Page
from mokr.browser.target import Target
from mokr.exceptions import BrowserError

if TYPE_CHECKING:
    from mokr.browser import Browser


LOGGER = logging.getLogger(__name__)


class BrowserContext(EventEmitter):
    def __init__(self, browser: Browser, context_id: str | None) -> None:
        """
        Browser context within the browser process. Contexts are independent
        of one another, not sharing storage. Contexts beyond the default
        context that `mokr.browser.Browser` starts with will be incognito.

        Args:
            browser (Browser): Parent `mokr.browser.Browser` object from which
                the context was spawned.
            context_id (str | None): The context identifier (may be None).
        """
        super().__init__()
        self._browser = browser
        self._id = context_id

    @property
    def incognito(self) -> bool:
        """
        True if the browser context is incognito, otherwise False. Only default
        context spawned when the `mokr.browser.Browser` object is spawned is
        not incognito.
        """
        return bool(self._id)

    @property
    def browser(self) -> Browser:
        """The `mokr.browser.Browser` object this context is attached to."""
        return self._browser

    def targets(self) -> list[Target]:
        """
        A list of all `mokr.browser.Target`s in this context.

        Returns:
            list[Target]: All initialised targets within this context.
        """
        return [
            target for target in self._browser.targets()
            if target.browser_context == self
        ]

    async def pages(self) -> list[Page]:
        """
        A list of all `mokr.browser.Page`s in this context.

        Returns:
            list[Page]: All pages within this context.
        """
        page_targets = [
            target for target in self.targets() if target.kind == "page"
        ]
        return [target._page for target in page_targets if await target.page()]

    async def first_page(self) -> Page:
        """
        Return the first page in `BrowserContext.pages`.
        If no pages active, returns None.

        Returns:
            Page | None: First active page in this context, if any.
        """
        pages = await self.pages()
        if pages:
            return pages[0]

    async def new_page(self) -> Page:
        """
        Spawn a new page.

        Returns:
            Page: A new `mokr.browser.Page` at "about:blank".
        """
        return await self._browser._create_page_in_context(self._id)

    async def close(self) -> None:
        """
        Close the context. Can only be done on incognito contexts.

        Raises:
            BrowserError: Raised if called from the default context.
        """
        if self._id is None:
            raise BrowserError('Non-incognito profile cannot be closed.')
        await self._browser._dispose_context(self._id)



import asyncio
from argparse import ArgumentParser

from mokr import launch
from mokr.constants import LIFECYCLE_EVENTS
from mokr.download import ensure_binary, install_binary


async def scrape(
    browser_type: str,
    url: str,
    headless: bool,
    timeout: int,
    wait_until: LIFECYCLE_EVENTS,
    user_agent: str | None,
    proxy: str | None,
    output_file: str | None,
) -> None:
    """
    Run a simple scrape of a URL.

    Args:
        browser_type (str): Either "chrome" or "firefox".
        url (str): Target URL to navigate to.
        headless (bool): Run in headles mode, if True.
        timeout (int): Tiem to wait in milliseconds for page load.
        wait_until (LIFECYCLE_EVENTS): Condition to wait for.
        user_agent (str | None): User agent to override page with.
        proxy (str | None): Remote proxy to use.
        output_file (str | None): Output file.

    Raises:
        SystemExit: Raised if any error encountered to set exit code to 1.
    """
    try:
        browser = await launch(
            browser_type=browser_type,
            headless=headless,
            proxy=proxy,
        )
        page = await browser.first_page()
        if user_agent:
            await page.set_user_agent(user_agent)
        response = await page.goto(url, timeout=timeout, wait_until=wait_until)
        if browser_type == "chrome":
            html = await response.content()
        else:
            html = (
                "Firefox does not support accessing response body."
                f" Response result: {response.status}: {response.reason}"
            )
        print(html)
        await browser.close()
        if output_file:
            with open(output_file, "w") as f:
                f.write(html)
    except Exception as error:
        print(f"Failed to scrape! Error: {error}")
        raise SystemExit(1)


def main() -> None:
    parser = ArgumentParser("mokr", description="A remote-controlled browser.")
    subparsers = parser.add_subparsers(dest="command")
    install_parser = subparsers.add_parser(
        "install",
        description="Install required browser."
    )
    install_parser.add_argument(
        "--force",
        help="Don't check first, just force install.",
        action="store_true",
    )
    install_parser.add_argument(
        "--type",
        help="Type of browser to install. If omitted, installs all.",
        choices=["chrome", "firefox"],
        dest="browser_type",
    )
    scrape_parser = subparsers.add_parser(
        "scrape",
        description="Run a browser session, navigate to a URL, and dump HTML.",
    )
    scrape_parser.add_argument(
        "url",
        help="Target URL to scrape.",
        type=str,
    )
    scrape_parser.add_argument(
        "--type",
        help="Type of browser to use.",
        choices=["chrome", "firefox"],
        dest="browser_type",
        default="chrome",
    )
    scrape_parser.add_argument(
        "--headful",
        help="Run in headful (not headless) browser.",
        action="store_true",
    )
    scrape_parser.add_argument(
        "--timeout",
        help="Time in milliseconds to wait for page to load.",
        type=int,
        default=30000,
    )
    scrape_parser.add_argument(
        "--wait-until",
        help="Wait until a specific lifecycle event occurs.",
        type=str,
        default="load",
    )
    scrape_parser.add_argument(
        "--user-agent",
        help="User agent to override the page with.",
        type=str,
        default=None,
    )
    scrape_parser.add_argument(
        "--proxy",
        help="Remote proxy to use.",
        type=str,
        default=None,
    )
    scrape_parser.add_argument(
        "--output-file",
        help="Dump HTML content to given file.",
        type=str,
        default=None,
    )
    options = parser.parse_args()
    if options.command == "install":
        browser_types = (
            [options.browser_type] if options.browser_type
            else ["chrome", "firefox"]
        )
        for browser_type in browser_types:
            if options.force or not ensure_binary(browser_type):
                install_binary(browser_type)
            else:
                print(f"{browser_type.title()} browser already installed.")
    elif options.command == "scrape":
        asyncio.run(
            scrape(
                options.browser_type,
                options.url,
                not options.headful,
                options.timeout,
                options.wait_until,
                options.user_agent,
                options.proxy,
                options.output_file,
            )
        )



CLOSE = "close"
DISCONNECTED = "disconnected"
ERROR = "error"
LIFECYCLE_EVENT = "lifecycleevent"
METRICS = "metrics"

BROWSER_CLOSE = "Browser.close"
BROWSER_GET_VERSION = "Browser.getVersion"

DOM_BOX_MODEL = "DOM.getBoxModel"
DOM_CONTENT_QUADS = "DOM.getContentQuads"
DOM_DESCRIBE_NODE = "DOM.describeNode"
DOM_INPUT_FILES = "DOM.setFileInputFiles"
DOM_LOADED = "domcontentloaded"

EMULATION_DISABLE_SCRIPT_EXECUTION = "Emulation.setScriptExecutionDisabled"
EMULATION_ENABLE_TOUCH = "Emulation.setTouchEmulationEnabled"
EMULATION_OVERRIDE_BACKGROUND = "Emulation.setDefaultBackgroundColorOverride"
EMULATION_OVERRIDE_METRICS = "Emulation.setDeviceMetricsOverride"
EMULATION_SET_EMULATED_MEDIA = "Emulation.setEmulatedMedia"

FETCH_AUTH_REQD = "Fetch.authRequired"
FETCH_CONTINUE = "Fetch.continueRequest"
FETCH_CONTINUE_AUTH = "Fetch.continueWithAuth"
FETCH_DISABLED = "Fetch.disable"
FETCH_DOMAIN_REQUEST_SENT = "FetchDomain.requestSent"
FETCH_ENABLED = "Fetch.enable"
FETCH_FAIL = "Fetch.failRequest"
FETCH_FULFILL = "Fetch.fulfillRequest"
FETCH_REQUEST_PAUSED = "Fetch.requestPaused"

FRAME_ATTACHED = "frameattached"
FRAME_DETACHED = "framedetached"
FRAME_NAVIGATED = "framenavigated"
FRAME_NAVIGATED_IN_DOC = "framenavigatedwithindocument"

INPUT_INSERT_TEXT = "Input.insertText"
INPUT_KEY = "Input.dispatchKeyEvent"
INPUT_MOUSE = "Input.dispatchMouseEvent"
INPUT_TOUCH = "Input.dispatchTouchEvent"

INSPECTOR_TARGET_CRASHED = "Inspector.targetCrashed"

LOG_ENABLE = "Log.enable"
LOG_ENTRY_ADDED = "Log.entryAdded"

NETWORK_CACHE_DISABLE = "Network.setCacheDisabled"
NETWORK_DELETE_COOKIES = "Network.deleteCookies"
NETWORK_EMULATE_NETWORK_CONDITIONS = "Network.emulateNetworkConditions"
NETWORK_ENABLE = "Network.enable"
NETWORK_EXTRA_HEADERS = "Network.setExtraHTTPHeaders"
NETWORK_GET_ALL_COOKIES = "Network.getAllCookies"
NETWORK_GET_COOKIES = "Network.getCookies"
NETWORK_GET_RESPONSE_BODY = "Network.getResponseBody"
NETWORK_LOADING_FAILED = "Network.loadingFailed"
NETWORK_LOADING_FINISHED = "Network.loadingFinished"
NETWORK_MGR_REQUEST = "NetworkManager.Request"
NETWORK_MGR_REQUEST_FAILED = "NetworkManager.RequestFailed"
NETWORK_MGR_REQUEST_FINISHED = "NetworkManager.RequestFinished"
NETWORK_MGR_REQUEST_FROM_CACHE = "NetworkManager.RequestServedFromCache"
NETWORK_MGR_RESPONSE = "NetworkManager.Response"
NETWORK_REQUEST = "request"
NETWORK_REQUEST_FAILED = "requestfailed"
NETWORK_REQUEST_FINISHED = "requestfinished"
NETWORK_REQUEST_SERVED_FROM_CACHE = "Network.requestServedFromCache"
NETWORK_REQUEST_WILL_BE_SENT = "Network.requestWillBeSent"
NETWORK_RESPONSE = "response"
NETWORK_RESPONSE_RECVD = "Network.responseReceived"
NETWORK_RESPONSE_RECVD_EXTRA = "Network.responseReceivedExtraInfo"
NETWORK_SET_COOKIES = "Network.setCookies"
NETWORK_USER_AGENT_OVERRIDE = "Network.setUserAgentOverride"

PAGE_ADD_SCRIPT_TO_EVAL = "Page.addScriptToEvaluateOnNewDocument"
PAGE_BRING_TO_FRONT = "Page.bringToFront"
PAGE_CLOSE = "Page.close"
PAGE_CONSOLE = "console"
PAGE_DIALOG = "dialog"
PAGE_DOM_LOADED = "Page.domContentEventFired"
PAGE_ENABLE = "Page.enable"
PAGE_ENABLE_LIFECYCLE_EVENTS = "Page.setLifecycleEventsEnabled"
PAGE_ERROR = "pageerror"
PAGE_FRAME_ATTACHED = "Page.frameAttached"
PAGE_FRAME_DETACHED = "Page.frameDetached"
PAGE_FRAME_NAVIGATED = "Page.frameNavigated"
PAGE_FRAME_NAVIGATED_IN_DOC = "Page.navigatedWithinDocument"
PAGE_FRAME_STOPPED_LOADING = "Page.frameStoppedLoading"
PAGE_GET_FRAME_TREE = "Page.getFrameTree"
PAGE_GET_LAYOUT = "Page.getLayoutMetrics"
PAGE_GET_NAVIGATION_HISTORY = "Page.getNavigationHistory"
PAGE_HANDLE_DIALOG = "Page.handleJavaScriptDialog"
PAGE_JAVASCRIPT_DIALOG_OPEN = "Page.javascriptDialogOpening"
PAGE_LIFECYCLE_EVENT = "Page.lifecycleEvent"
PAGE_LOAD = "load"
PAGE_LOAD_EVENT_FIRED = "Page.loadEventFired"
PAGE_NAVIGATE = "Page.navigate"
PAGE_NAVIGATE_TO_HISTORY_ENTRY = "Page.navigateToHistoryEntry"
PAGE_PRINT_TO_PDF = "Page.printToPDF"
PAGE_RELOAD = "Page.reload"
PAGE_SCREENSHOT = "Page.captureScreenshot"
PAGE_SET_BYPASS_CSP = "Page.setBypassCSP"

PERFORMANCE_ENABLE = "Performance.enable"
PERFORMANCE_GET_METRICS = "Performance.getMetrics"
PERFORMANCE_METRICS = "Performance.metrics"

RUMTIME_BINDING_CALL = "Runtime.bindingCalled"
RUNTIME_ADD_BINDING = "Runtime.addBinding"
RUNTIME_CALL_FUNCTION = "Runtime.callFunctionOn"
RUNTIME_CONSOLE_API_CALL = "Runtime.consoleAPICalled"
RUNTIME_ENABLE = "Runtime.enable"
RUNTIME_EVALUATE = "Runtime.evaluate"
RUNTIME_EXCEPTION_THROWN = "Runtime.exceptionThrown"
RUNTIME_EXECUTION_CONTEXTS_CLEARED = "Runtime.executionContextsCleared"
RUNTIME_EXECUTION_CONTEXT_CREATED = "Runtime.executionContextCreated"
RUNTIME_EXECUTION_CONTEXT_DESTROYED = "Runtime.executionContextDestroyed"
RUNTIME_GET_PROPERTIES = "Runtime.getProperties"
RUNTIME_QUERY_OBJECTS = "Runtime.queryObjects"
RUNTIME_RELEASE_OBJECT = "Runtime.releaseObject"

SECURITY_ENABLE = "Security.enable"
SECURITY_IGNORE_HTTPS_ERRORS = "Security.setIgnoreCertificateErrors"

TARGET_ACTIVATE = "Target.activateTarget"
TARGET_ATTACH = "Target.attachToTarget"
TARGET_ATTACHED = "Target.attachedToTarget"
TARGET_CHANGED = "targetchanged"
TARGET_CLOSE = "Target.closeTarget"
TARGET_CREATED = "targetcreated"
TARGET_CREATE_BROWSER_CONTEXT = "Target.createBrowserContext"
TARGET_CREATE_TARGET = "Target.createTarget"
TARGET_DESTROYED = "targetdestroyed"
TARGET_DETACHED = "Target.detachedFromTarget"
TARGET_DISPOSE_BROWSER_CONTEXT = "Target.disposeBrowserContext"
TARGET_GET_CONTEXTS = "Target.getBrowserContexts"
TARGET_INFO_CHANGED = "Target.targetInfoChanged"
TARGET_RECV_MSG = "Target.receivedMessageFromTarget"
TARGET_SEND_DETACH = "Target.detachFromTarget"
TARGET_SEND_MSG = "Target.sendMessageToTarget"
TARGET_SET_AUTO_ATTACH = "Target.setAutoAttach"
TARGET_SET_DISCOVER_TARGETS = "Target.setDiscoverTargets"
TARGET_TARGET_CREATED = "Target.targetCreated"
TARGET_TARGET_DESTROYED = "Target.targetDestroyed"

WORKER_CREATED = "workercreated"
WORKER_DESTROYED = "workerdestroyed"



import logging
import math
from typing import Any, Awaitable, Callable

from pyee import EventEmitter

from mokr.connection import DevtoolsConnection
from mokr.constants import RUNTIME_RELEASE_OBJECT
from mokr.exceptions import ElementHandleError


LOGGER = logging.getLogger(__name__)


def format_javascript_exception(exception_details: dict) -> str:
    """
    Get and format a JavaScript exception message.

    Args:
        exception_details (dict): An "exceptionDetails" object loaded as
            a dictionary.

    Returns:
        str: Formatted exception as string.
    """
    exception = exception_details.get('exception')
    if exception:
        return exception.get('description') or exception.get('value')
    message = exception_details.get('text', '')
    stackTrace = exception_details.get('stackTrace', dict())
    if stackTrace:
        for callframe in stackTrace.get('callFrames'):
            location = (
                f"{callframe.get('url', '')}:"
                f"{callframe.get('lineNumber', '')}:"
                f"{callframe.get('columnNumber')}"
            )
            functionName = callframe.get('functionName', '<anonymous>')
            message = message + f'\n    at {functionName} ({location})'
    return message


def add_event_listener(
    emitter: EventEmitter,
    event_name: str,
    handler: Callable,
) -> dict[str, Any]:
    """
    Add an event `handler` to an `emitter` for a given `event_name`.

    Args:
        emitter (EventEmitter): Emitter to listen on.
        event_name (str): Event name to listen for.
        handler (Callable): Handler callback to run when event is emitted.

    Returns:
        dict[str, Any]: A dictionary representation of the listener.
    """
    emitter.on(event_name, handler)
    return {'emitter': emitter, 'eventName': event_name, 'handler': handler}


def remove_event_listeners(listeners: list[dict]) -> None:
    """
    Remove all event listeners given from their emitters.

    Args:
        listeners (list[dict]): A list of listeners in dictionary form, as
            returned by `mokr.utils.remote.add_event_listener`.
    """
    for listener in listeners:
        emitter = listener['emitter']
        event_name = listener['eventName']
        handler = listener['handler']
        if event_name in emitter._events.keys():
            try:
                emitter.remove_listener(event_name, handler)
            except KeyError:
                # Ignore keyerrors, event isn't registered.
                pass
    listeners.clear()


def serialize_remote_object(remote_object: dict) -> Any:
    """
    Transform a remote JavaScript object's "value" to a Python equivalent.

    Args:
        remote_object (dict): The remote JavaScript object to serialise.

    Raises:
        ElementHandleError: Raised if the `remote_object` has an "objectId"
        (is not a primitive object) or if an unhandled case is encountered
        when checking `remote_object["unserializableValue"]`.

    Returns:
        Any: Python representation of remote object. May be str, int, dict,
            list, etc. Some `remote_object["unserializableValue"]`s will be
            transformed into equivalents including "Nan" to None, "-0" to 0,
            "Infinity" to `math.inf`, and "-Infinity" to `-math.inf`.
    """
    if remote_object.get('objectId'):
        raise ElementHandleError('Cannot extract value when objectId is given.')
    value = remote_object.get('unserializableValue')
    if value:
        if value == '-0':
            return -0
        elif value == 'NaN':
            return None
        elif value == 'Infinity':
            return math.inf
        elif value == '-Infinity':
            return -math.inf
        else:
            raise ElementHandleError(f'Unserializable value: {value}')
    return remote_object.get('value')


def release_remote_object(
    client: DevtoolsConnection,
    remote_object: dict
) -> Awaitable:
    """
    Release a given `remote_object` so that it is no longer referenced
    by the browser and can be garbage collected.

    Ignores all exceptions raised when sending request to devtools session.

    Args:
        client (DevtoolsConnection): A `mokr.connection.DevtoolsConnection`.
        remote_object (dict): A remote object as dictionary.

    Returns:
        Awaitable: Awaitable that yields None.
    """
    object_id = remote_object.get('objectId')
    fut_none = client._loop.create_future()
    fut_none.set_result(None)
    if not object_id:
        return fut_none
    try:
        return client.send(
            RUNTIME_RELEASE_OBJECT,
            {'objectId': object_id}
        )
    except Exception:
        # Harmless exceptions may happen if page navigated or closed.
        LOGGER.debug(
            "Ignoring exception releasing remote object.",
            exc_info=True,
        )
    return fut_none


def is_javascript_method(method: str) -> bool:
    """
    Casually check if string is a JavaScript method.

    Args:
        method (str): String to check.

    Returns:
        bool: True if deemed to be a method, else False.
    """
    method = method.strip()
    if (
        method.startswith('function')
        or method.startswith('async ')
        or '=>' in method
    ):
        return True
    return False



from __future__ import annotations

import base64
import copy
import inspect
import logging
from typing import TYPE_CHECKING, Callable

import httpx

from mokr.connection import DevtoolsConnection
from mokr.constants import (
    FETCH_CONTINUE,
    FETCH_FAIL,
    FETCH_FULFILL,
    HTTP_RESPONSE_CODES_TEXT_MAP,
    NETWORK_ERROR_CODES_TO_REASONS,
)
from mokr.exceptions import NetworkError
from mokr.frame import Frame

if TYPE_CHECKING:
    from mokr.browser.page import Page
    from mokr.network.response import Response


LOGGER = logging.getLogger(__name__)


class Request():
    def __init__(
        self,
        page: Page,
        client: DevtoolsConnection,
        request_id: str | None,
        interception_id: str | None,
        is_navigation_request: bool,
        allow_interception: bool,
        url: str,
        resource_type: str,
        payload: dict,
        frame: Frame | None,
        redirect_chain: list[Request],
        interception_callback_chain: list[Callable],
        httpx_request: httpx.Request | None = None,
    ) -> None:
        """
        Representative of a remote network request, this object is created
        by a `mokr.network.NetworkManager`.

        The creation of a `Request` does not signify the request has been sent,
        or that a request response has been received.

        Args:
            page (Page): Parent `mokr.browser.Page`.
            client (DevtoolsConnection): A `mokr.connection.DevtoolsConnection`
                spawned by the parent `mokr.network.NetworkManager`.
            request_id (str | None): The network request identifier, from the
                remote request event's requestId.
            interception_id (str | None): The fetch request identifier, from the
                remote request event's interceptionId. This can be confusing as
                the fetch domain refers to this as the requestId.
            is_navigation_request (bool): Whether or not this request will
                affect remote frame navigation.
            allow_interception (bool): Whether or not to allow "request" event
                interception.
            url (str): The remote request URL.
            resource_type (str): Resource type as it was perceived by the
                rendering engine. See `Request.resource_type` for more.
            payload (dict): Request data and metadata.
            frame (Frame | None): The `mokr.frame.Frame` this request originated
                from.
            redirect_chain (list[Request]): A list of `Request` objects that
                were redirects and lead to this request.
            interception_callback_chain (list[Callable]): A list of callbacks
                shared with the parent `mokr.network.NetworkManager` and its
                parent `mokr.browser.Page`. These callbacks wil run sequentially
                during "request" event interception. See `mokr.browser.Page` for
                more information.
            httpx_request: (httpx.Request | None, optional): If created by a
                `mokr.network.HttpDomain`, the original `httpx.Request` that
                this will be built from.
        """
        self._page = page
        self._client = client
        self._request_id = request_id
        self._is_navigation_request = is_navigation_request
        self._interception_id = interception_id
        self._allow_interception = allow_interception
        self._interception_handled = False
        self._response: Response | None = None
        self._failure_text: str | None = None
        self._url = url
        self._resource_type = resource_type.lower()
        self._method = payload.get('method')
        self._post_data = payload.get('postData')
        headers = payload.get('headers', {})
        self._headers = {k.lower(): v for k, v in headers.items()}
        self._frame = frame
        self._redirect_chain = redirect_chain
        self._from_memory_cache = False
        self._interception_callback_chain = interception_callback_chain
        # This will only be set in FetchDomain handling.
        self._mokr_request_uuid = None
        self._httpx_request = httpx_request

    def __repr__(self) -> str:
        return f"<Request ({self.method} to {self.url}) at {id(self)}>"

    @property
    def url(self) -> str:
        """URL from the remote request."""
        return self._url

    @property
    def resource_type(self) -> str:
        """
        Resource type as it was perceived by the rendering engine.

        One of "Document", "Stylesheet", "Image", "Media", "Font", "Script",
        "TextTrack", "XHR", "Fetch", "Prefetch", "EventSource", "WebSocket",
        "Manifest", "SignedExchange", "Ping", "CSPViolationReport", "Preflight",
        or "Other".
        """
        return self._resource_type

    @property
    def method(self) -> str | None:
        """
        HTTP request method from remote request, such as "GET", "POST", "PATCH".
        """
        return self._method

    @property
    def post_data(self) -> str | None:
        """Data from this `Request`'s `payload["postData"]`, if any."""
        return self._post_data

    @property
    def headers(self) -> dict:
        """All headers for the target request with keys lowered."""
        return self._headers

    @property
    def response(self) -> Response | None:
        """
        The `mokr.network.Response` object created from the remote response
        bound to the remote request, or None if not yet received.
        """
        return self._response

    @property
    def frame(self) -> Frame | None:
        """The corresponding `mokr.frame.Frame`, if any."""
        return self._frame

    @property
    def redirect_chain(self) -> list['Request']:
        """
        Return the chain of `Request` objects if the first remote request
        was a redirect.

        The chain will be shared will all subsequent `Request` objects initiated
        from the redirect, and the final `Request` in the chain will not be
        a redirect.
        """
        return copy.copy(self._redirect_chain)

    @property
    def httpx_request(self) -> bool:
        """
        If created by a `mokr.network.HttpDomain`, this will hold the original
        `httpx.Request` object this was populated from. Otherwise, None.
        """
        return self._httpx_request

    def _verify_request_allowed(self) -> None:
        if not self._allow_interception:
            raise NetworkError('Request interception is not enabled.')
        if self._interception_handled:
            raise NetworkError('Request is already handled.')
        if self._httpx_request:
            raise ValueError(
                "Cannot run interception methods on HttpDomain-based requests."
            )

    async def _run_callback_chain(self, callback_chain: list[Callable]) -> None:
        result = None
        callback_chain = callback_chain.copy()
        for index, callback in enumerate(callback_chain):
            if index == 0:
                args = [self]
            else:
                if result:
                    args = [result]
                else:
                    return
            if inspect.iscoroutinefunction(callback):
                result = await callback(*args)
            else:
                result = callback(*args)

    async def _default_request_intercept(self, request: Request) -> None:
        if not self._page._is_firefox(mute=True):
            return await request.release()

    async def _finalize_interceptions(self) -> None:
        if not self._allow_interception:
            return
        self._verify_request_allowed()
        # If somehow the default intercept the `Page` added was removed and
        # the chain is completely empty, use a similar default here.
        if not self._interception_callback_chain:
            self._interception_callback_chain.append(
                self._default_request_intercept
            )
        await self._run_callback_chain(self._interception_callback_chain)

    def is_navigation_request(self) -> bool:
        """
        Whether or not this request will affect remote frame navigation.

        Returns:
            bool: True if navigating the frame, otherwise False.
        """
        return self._is_navigation_request

    def failure_text(self) -> str | None:
        """
        Return error text from failed requests. For successful requests, this
        will return None.

        Returns:
            str | None: Error text if request failed, otherwise None.
        """
        return self._failure_text

    async def release(
        self,
        url: str = None,
        method: str = None,
        post_data: str = None,
        headers: dict = None,
    ) -> None:
        """
        Release an intercepted request, optionally altering select parameters.

        Stops execution of the rest of the interception chain.

        Args:
            url (str, optional): The URL to overwrite the request with, if any.
                Defaults to None (use initiated `Request.url`).
            method (str, optional): The HTTP request method (e.g. "GET", "POST")
                to overwrite the request with, if any.
                Defaults to None (use initiated `Request.url`).
            post_data (str, optional): The post data (payload["postData"]) to
                overwrite the request with, if any.
                Defaults to None (use initiated `Request.post_data`).
            headers (dict, optional): Headers, will completely replace existing
                headers. Defaults to None (use `Request.headers`).
        """
        if self._url.startswith('data:'):
            return
        self._verify_request_allowed()
        self._interception_handled = True
        # Sometimes we have no fetch ID (interception ID), but the request ID
        # is usually the interception ID in these edge-cases.
        if not self._interception_id:
            self._interception_id = self._request_id
        params = {'requestId': self._interception_id}
        overrides = {
            k: v for k, v in {
                "url": url,
                "method": method,
                "postData": post_data,
                "headers": headers,
            }.items()
            if v is not None
        }
        params.update(overrides)
        try:
            await self._client.send(FETCH_CONTINUE, params)
        except Exception:
            LOGGER.error("Error continuing network request.", exc_info=True)

    async def fulfill(
        self,
        response: Response | None = None,
        status: int = 200,
        headers: dict | None = None,
        body: str | bytes | None = None,
    ) -> None:
        """
        Respond to a request with the given response data. Prevents the
        remote request from being actually sent. Cannot be used with
        data urls.

        Stops execution of the rest of the interception chain.

        Args:
            response (Response | None, optional): A completed
                `mokr.network.Reponse` object to extract data from. Passing
                this will ignore all other kwargs passed.
                Defaults to None.
            status (int, optional): HTTP status code. Defaults to 200.
            headers (dict | None, optional): Response headers.
                Defaults to None (empty headers).
            body (str | bytes | None, optional): Response body.
                Defaults to None (empty body).
        """
        if self._url.startswith('data:'):
            return
        self._verify_request_allowed()
        if response:
            body = await response.content()
            status = response.status
            headers = response.headers
        overrides = {
            "responseCode": status,
            "responsePhrase": HTTP_RESPONSE_CODES_TEXT_MAP[str(status)],
            "responseHeaders": headers,
            "body": body,
        }
        self._interception_handled = True
        fields = ["responseCode", "responsePhrase", "responseHeaders", "body"]
        overrides = {
            k: v for k, v in overrides.items()
            if v is not None and k in fields
        }
        headers_dict = {}
        if overrides.get('responseHeaders'):
            for name, value in overrides['responseHeaders'].items():
                headers_dict[name.lower()] = value
        if overrides.get('body'):
            if isinstance(overrides['body'], str):
                body = body.encode()
            if 'content-length' not in headers_dict:
                headers_dict['content-length'] = len(body)
            overrides["body"] = base64.b64encode(body).decode()
        prepared_headers = []
        if headers_dict:
            for name, value in headers_dict.items():
                prepared_headers.append({"name": name, "value": str(value)})
        overrides["responseHeaders"] = prepared_headers
        params = {'requestId': self._interception_id, **overrides}
        try:
            await self._client.send(FETCH_FULFILL, params)
        except Exception:
            LOGGER.error("Error fulfilling request.", exc_info=True)

    async def abort(self, error_reason: str = "failed") -> None:
        """
        Abort the remote request. This will prevent the request from being sent,
        which can cause a `mokr.exceptions.PageError` to raise later if
        `Request.is_navigation_request` is True.

        Stops execution of the rest of the interception chain.

        Args:
            error_reason (str, optional): The error reason to give when aborting
                the request. One of "Failed", "Aborted", "TimedOut",
                "AccessDenied", "ConnectionClosed", "ConnectionReset",
                "ConnectionRefused", "ConnectionAborted", "ConnectionFailed",
                "NameNotResolved", "InternetDisconnected", "AddressUnreachable",
                "BlockedByClient", or "BlockedByResponse".
                Defaults to "failed".

        Raises:
            NetworkError: Raised if unknown error reason is given.
        """
        if error_reason not in NETWORK_ERROR_CODES_TO_REASONS.values():
            error_reason = NETWORK_ERROR_CODES_TO_REASONS[error_reason]
            if not error_reason:
                raise NetworkError(f'Unknown error reason: {error_reason}')
        self._verify_request_allowed()
        self._interception_handled = True
        try:
            await self._client.send(
                FETCH_FAIL,
                {
                    "requestId": self._interception_id,
                    "errorReason": error_reason,
                },
            )
        except Exception:
            LOGGER.error("Error aborting request.", exc_info=True)



NETWORK_ERROR_CODES_TO_REASONS = {
    'aborted': 'Aborted',
    'accessdenied': 'AccessDenied',
    'addressunreachable': 'AddressUnreachable',
    'blockedbyclient': 'BlockedByClient',
    'blockedbyresponse': 'BlockedByResponse',
    'connectionaborted': 'ConnectionAborted',
    'connectionclosed': 'ConnectionClosed',
    'connectionfailed': 'ConnectionFailed',
    'connectionrefused': 'ConnectionRefused',
    'connectionreset': 'ConnectionReset',
    'internetdisconnected': 'InternetDisconnected',
    'namenotresolved': 'NameNotResolved',
    'timedout': 'TimedOut',
    'failed': 'Failed',
}

HTTP_RESPONSE_CODES_TEXT_MAP = {
    '100': 'Continue',
    '101': 'Switching Protocols',
    '102': 'Processing',
    '200': 'OK',
    '201': 'Created',
    '202': 'Accepted',
    '203': 'Non-Authoritative Information',
    '204': 'No Content',
    '206': 'Partial Content',
    '207': 'Multi-Status',
    '208': 'Already Reported',
    '209': 'IM Used',
    '300': 'Multiple Choices',
    '301': 'Moved Permanently',
    '302': 'Found',
    '303': 'See Other',
    '304': 'Not Modified',
    '305': 'Use Proxy',
    '306': 'Switch Proxy',
    '307': 'Temporary Redirect',
    '308': 'Permanent Redirect',
    '400': 'Bad Request',
    '401': 'Unauthorized',
    '402': 'Payment Required',
    '403': 'Forbidden',
    '404': 'Not Found',
    '405': 'Method Not Allowed',
    '406': 'Not Acceptable',
    '407': 'Proxy Authentication Required',
    '408': 'Request Timeout',
    '409': 'Conflict',
    '410': 'Gone',
    '411': 'Length Required',
    '412': 'Precondition Failed',
    '413': 'Payload Too Large',
    '414': 'URI Too Long',
    '415': 'Unsupported Media Type',
    '416': 'Range Not Satisfiable',
    '417': 'Expectation Failed',
    '418': 'I\'m a teapot',
    '421': 'Misdirected Request',
    '422': 'Unprocessable Entity',
    '423': 'Locked',
    '424': 'Failed Dependency',
    '426': 'Upgrade Required',
    '428': 'Precondition Required',
    '429': 'Too Many Requests',
    '431': 'Request Header Fields Too Large',
    '451': 'Unavailable For Legal Reasons',
    '500': 'Internal Server Error',
    '501': 'Not Implemented',
    '502': 'Bad Gateway',
    '503': 'Service Unavailable',
    '504': 'Gateway Timeout',
    '505': 'HTTP Version Not Supported',
    '506': 'Variant Also Negotiates',
    '507': 'Insufficient Storage',
    '508': 'Loop Detected',
    '510': 'Not Extended',
    '511': 'Network Authentication Required',
}



import os
import shutil
import tempfile
import time
from copy import copy
from urllib.parse import urlparse

from mokr.constants import INSTALL_PATH
from mokr.launch.base import Launcher


CHROME_PROFILE_PATH = INSTALL_PATH / '.dev_profile'

DEFAULT_CHROME_ARGS = [
    '--allow-pre-commit-input',
    '--disable-background-networking',
    '--disable-background-timer-throttling',
    '--disable-background-occluded-windows',
    '--disable-breakpad',
    '--disable-client-side-phishing-detection',
    '--disable-component-extensions-with-background-pages',
    '--disable-component-update',
    '--disable-default-apps',
    '--disable-dev-shm-usage',
    '--disable-extensions',
    '--disable-field-trial-config',
    '--disable-infobars',
    '--disable-ipc-flooding-protection',
    '--disable-popup-blocking',
    '--disable-prompt-on-repost',
    '--disable-renderer-backgrounding',
    '--disable-search-engine-choice-screen',
    '--disable-sync',
    '--enable-automation',
    '--export-tagged-pdf',
    '--generate-pdf-document-outline',
    '--force-color-profile=srgb',
    '--metrics-recording-only',
    '--no-first-run',
    '--password-store=basic',
    '--use-mock-keychain',
    '--enable-features=NetworkServiceInProcess2',
    '--disable-features=Translate,AcceptCHFrame,MediaRouter,OptimizationHints,ProcessPerSiteUpToMainFrameThreshold',  # noqa
    # '--disable-browser-side-navigation',
    # '--safebrowsing-disable-auto-update',
    # '--disable-hang-monitor',
    # '--disable-translate',
]


class ChromeLauncher(Launcher):
    kind = "chrome"

    def _parse_proxy(self, proxy: str) -> None:
        proxy_parts = urlparse(proxy)
        scheme = proxy_parts.scheme
        host = proxy_parts.hostname
        port = proxy_parts.port
        if proxy.startswith("socks"):
            args = [
                f'--proxy-server={proxy}',
                f'--host-resolver-rules="MAP * ~NOTFOUND , EXCLUDE {host}"',
            ]
        else:
            proxy = f"{scheme}://{host}:{port}"
            args = [f'--proxy-server={proxy}']
        self.browser_arguments.extend(args)
        credentials = {}
        credentials["proxy"] = proxy
        if proxy_parts.username:
            credentials["username"] = proxy_parts.username
        if proxy_parts.password:
            credentials["password"] = proxy_parts.password
        return credentials

    def _custom_default_args(
        self,
        headless: bool = None,
        user_data_dir: str = None,
        devtools: bool = False,
    ):
        browser_arguments = copy(DEFAULT_CHROME_ARGS)
        if user_data_dir:
            browser_arguments.append(f'--user-data-dir={user_data_dir}')
        if devtools:
            browser_arguments.append('--auto-open-devtools-for-tabs')
        if headless:
            browser_arguments.extend(
                ('--headless=new', '--hide-scrollbars', '--mute-audio')
            )

        return browser_arguments

    def _compute_launch_args(
        self,
        headless: bool = None,
        user_data_dir: str = None,
        devtools: bool = False,
        ignore_default_args: bool | list[str] = False,
        proxy: str = None,
        args: list[str] = None,
    ):
        super()._compute_launch_args(
            headless=headless,
            user_data_dir=user_data_dir,
            devtools=devtools,
            ignore_default_args=ignore_default_args,
            proxy=proxy,
            args=args,
        )
        self.proxy_credentials = self._parse_proxy(proxy) if proxy else None
        self.temp_user_data_dir = None
        if not any(
            arg for arg in self.browser_arguments
            if arg.startswith('--remote-debugging-')
        ):
            self.browser_arguments.append(
                f'--remote-debugging-port={self.port}'
            )
        if not any(
            arg for arg in self.browser_arguments
            if arg.startswith('--user-data-dir')
        ):
            if not CHROME_PROFILE_PATH.exists():
                CHROME_PROFILE_PATH.mkdir(parents=True)
            self.temp_user_data_dir = tempfile.mkdtemp(
                dir=str(CHROME_PROFILE_PATH)
            )
            self.browser_arguments.append(
                f'--user-data-dir={self.temp_user_data_dir}'
            )

    def _clean_restore_data_dirs(self) -> None:
        for _ in range(100):
            if self.temp_user_data_dir and os.path.exists(
                self.temp_user_data_dir
            ):
                shutil.rmtree(self.temp_user_data_dir, ignore_errors=True)
                if os.path.exists(self.temp_user_data_dir):
                    time.sleep(0.01)
            else:
                return
        else:
            raise IOError(
                'Unable to remove temporary user data dir'
                f' at {self.temp_user_data_dir}'
            )



import logging


LOGGER = logging.getLogger(__name__)


class SecurityDetails():
    def __init__(
        self,
        subject_name: str,
        issuer: str,
        valid_from: int,
        valid_to: int,
        protocol: str,
    ) -> None:
        """
        A simple representation of remote response security details.

        Args:
            subject_name (str): Subject to which the certificate was issued to.
            issuer (str): Name of issuer of the certificate.
            valid_from (int): Unix time string of start of certificate validity.
            valid_to (int): Unix time string of end of certificate validity.
            protocol (str): Security protocol (e.g. "TLS1.2", "TLS1.3").
        """
        self._subject_name = subject_name
        self._issuer = issuer
        self._valid_from = valid_from
        self._valid_to = valid_to
        self._protocol = protocol

    @property
    def subject_name(self) -> str:
        """Subject to which the certificate was issued to."""
        return self._subject_name

    @property
    def issuer(self) -> str:
        """Name of issuer of the certificate."""
        return self._issuer

    @property
    def valid_from(self) -> int:
        """Unix time string of start of certificate validity."""
        return self._valid_from

    @property
    def valid_to(self) -> int:
        """Unix time string of end of certificate validity."""
        return self._valid_to

    @property
    def protocol(self) -> str:
        """Security protocol (e.g. "TLS1.2", "TLS1.3")."""
        return self._protocol



import asyncio
import inspect
from typing import Any, Awaitable, Callable

from pyee import EventEmitter

from mokr.exceptions import MokrTimeoutError
from mokr.utils.remote import add_event_listener, remove_event_listeners


class EventWaiter():
    def __init__(
        self,
        emitter: EventEmitter,
        event_name: str,
        predicate: Callable,
        timeout: float,
        loop: asyncio.AbstractEventLoop,
    ) -> None:
        """
        Class to wait for an event to be emitted and then run a callback
        `predicate` when it is. If the callback resolves to a truthy value,
        the timer is cancelled. Otherwise, a `mokr.exceptions.MokrTimeoutError`
        is raised.

        Args:
            emitter (EventEmitter): Emitter to listen to.
            event_name (str): Event name to listen for.
            predicate (Callable): Callback to run when event is emitted.
            timeout (float): Time in milliseconds to wait for event.
            loop (asyncio.AbstractEventLoop): Running asyncio loop.
        """
        self._emitter = emitter
        self._event_name = event_name
        self._predicate = predicate
        self._timeout = timeout
        self._loop = loop
        self._event_timeout = None
        self.listener = None
        self.promise = loop.create_future()

    def _resolve_callback(self, target: Any) -> None:
        if not self.promise.done():
            self.promise.set_result(target)

    def _reject_callback(self, exception: Exception) -> None:
        self.promise.set_exception(exception)

    def _cleanup(self) -> None:
        remove_event_listeners([self.listener])
        if self._event_timeout:
            self._event_timeout.cancel()

    def _listener(self, target: Any) -> None:
        if not self._predicate(target):
            return
        self._cleanup()
        self._resolve_callback(target)

    async def _alistener(self, target: Any) -> None:
        if not await self._predicate(target):
            return
        self._cleanup()
        self._resolve_callback(target)

    async def _timeout_timer(self) -> None:
        await asyncio.sleep(self._timeout / 1000)
        self._reject_callback(
            MokrTimeoutError('Timeout exceeded while waiting for event.')
        )

    def wait(self) -> Awaitable:
        """
        Wait for the target event to be emitted and then run the target
        predicate when it is. If that resolves to a truthy value, cancel the
        timer, otherwise, raise a `mokr.exceptions.MokrTimeoutError` exception.

        Returns:
            Awaitable: Awaitable that yields result of the target callback.
        """
        self.listener = add_event_listener(
            self._emitter,
            self._event_name,
            (
                lambda target: self._loop.create_task(self._alistener(target))
                if inspect.iscoroutinefunction(self._predicate)
                else self._listener
            ),
        )
        if self._timeout:
            self._event_timeout = self._loop.create_task(self._timeout_timer())
        return self.promise



from mokr.network.event import NetworkEventManager  # noqa
from mokr.network.fetch import FetchDomain  # noqa
from mokr.network.http import HttpDomain  # noqa
from mokr.network.manager.base import NetworkManager  # noqa
from mokr.network.manager.chrome import ChromeNetworkManager  # noqa
from mokr.network.manager.firefox import FirefoxNetworkManager  # noqa
from mokr.network.request import Request  # noqa
from mokr.network.response import Response  # noqa
from mokr.network.security import SecurityDetails  # noqa



import asyncio
import gc
import logging
import socket
import subprocess
from abc import ABC

from mokr.browser.browser import Browser
from mokr.browser.target import Target
from mokr.connection import Connection
from mokr.constants import BROWSER_CLOSE
from mokr.download import browser_binary, ensure_binary
from mokr.utils import (
    add_event_listener,
    get_ws_endpoint,
    remove_event_listeners,
)


LOGGER = logging.getLogger(__name__)


class Launcher(ABC):
    kind = "abstract"

    def __init__(
        self,
        binary_path: str = None,
        headless: bool = None,
        user_data_dir: str = None,
        devtools: bool = False,
        ignore_default_args: bool | list[str] = False,
        ignore_https_errors: bool = False,
        default_viewport: dict[str, int] = None,
        proxy: str = None,
        default_user_agent: str = None,
        slow_mo: int = 0,
        log_level: str | int = None,
        args: list[str] = None,
        dumpio: bool = False,
        env: dict[str, str] = None,
        loop: asyncio.AbstractEventLoop = None,
        firefox_user_prefs: dict = None,
        firefox_addons_paths: list[str] = None,
    ) -> None:
        """
        Class to handle launching browser process and creation of a
        `mokr.browser.Browser` object.

        Args:
            binary_path (str, optional): Path to executable to use.
                Defaults to None (looks for default executable that can be
                installed via `mokr install`).
            headless (bool, optional): Run the browser in headless (no window)
                mode. Defaults to None (uses opposite value to `devtools`).
            user_data_dir (str, optional): Path to a user data directory.
                Defaults to None.
            devtools (bool, optional): Automatically open the developer tools
                panel. Defaults to False.
            ignore_default_args (bool | list[str], optional): Either a bool to
                indicate ignoring all arguments or a list of arguments to
                ignore. Be cautious, ignoring some arguments may cause
                unexpected results.
                Defaults to False.
            ignore_https_errors (bool, optional): Ignore site security errors.
                Defaults to False.
            default_viewport (dict[str, int], optional): Set the default
                viewport for new pages. Accepts a dictionary keyed with viewport
                options. Not all viewport options are considered, only:
                "isMobile", "width", "height", "deviceScaleFactor",
                "isLandscape", and "hasTouch".
                Defaults to None (800x600 viewport).
            proxy (str, optional): Proxy to route all requests through. Can
                be a regular HTTP/S proxy or SOCKS proxy. Expects proxy as
                `<scheme>://[(optional)<username>:<password>]@<host><password>`.
            default_user_agent (str, optional): Default user agent to use on
                all new pages.
            slow_mo (int, optional): Slow execution of remote calls by the given
                time in milliseconds. Defaults to 0.
            log_level (str | int, optional): Log level to log at.
                Defaults to None (same as root).
            args (list[str], optional): Additional arguments to pass to the
                browser process when launching. Defaults to None.
            dumpio (bool, optional): Pipe the browser process' stdout and stderr
                into ``process.stdout`` and ``process.stderr``.
                Defaults to False.
            env (dict[str, str], optional): Additional environment variables
                that the browser process will be able to read. Defaults to None.
            loop (asyncio.AbstractEventLoop, optional): A running asyncio loop
                to execute within. Defaults to None (uses
                `asyncio.get_event_loop`).
            firefox_user_prefs (dict): Firefox only. User preferences to load.
            firefox_addons_paths (list[str]): Firefox only. A list of paths to
                addons that will be installed as temporary extensions.

        Example::

            ```python
            from mokr.launch import ChromeLauncher, FirefoxLauncher

            async with ChromeLauncher() as browser:
                page = await browser.first_page()
                await page.goto("https://example.com")

            # Or, to avoid the contextmanager.
            launcher = FirefoxLauncher()
            browser = await launcher.launch()
            page = await browser.first_page()
            await page.goto("https://example.com")
            await launcher.stop()
            ```
        """
        self.port = self._get_free_port()
        self.url = f'http://127.0.0.1:{self.port}'
        self._loop = loop if loop else asyncio.get_event_loop()
        self.dumpio = dumpio
        self.env = env
        self.ignore_https_errors = ignore_https_errors
        self.default_viewport = (
            default_viewport if default_viewport
            else {'width': 800, 'height': 600}
        )
        self.default_user_agent = default_user_agent
        self.slow_mo = slow_mo
        self.firefox_user_prefs = firefox_user_prefs
        self.firefox_addons_paths = firefox_addons_paths
        if log_level is not None:
            logging.getLogger('mokr').setLevel(log_level)
        self.browser_closed = True
        self.browser_arguments: list[str] = list()
        args = args if args else []
        self._compute_launch_args(
            headless,
            user_data_dir,
            devtools,
            ignore_default_args,
            proxy,
            args,
        )
        self.browser_binary = binary_path
        browser_type = getattr(self, "kind", "")
        if not self.browser_binary:
            binary = str(browser_binary(browser_type))
            if not ensure_binary(browser_type):
                self._print_needs_install_message(binary)
            self.browser_binary = binary
        self.cmd = [self.browser_binary] + self.browser_arguments
        self.initial_page_promise = self._loop.create_future()

    async def __aenter__(self, *args, **kwargs) -> Browser:
        browser = await self.launch()
        return browser

    async def __aexit__(self, *args, **kwargs) -> None:
        await self.stop()

    def _default_args(
        self,
        headless: bool = None,
        args: list[str] = None,
        user_data_dir: str = None,
        devtools: bool = False,
    ) -> list[str]:
        if headless is None:
            headless = not devtools
        args = args if args else []
        browser_arguments = self._custom_default_args(
            headless,
            user_data_dir,
            devtools,
        )
        browser_arguments.extend(args)
        if all(arg.startswith("-") for arg in args):
            browser_arguments.append('about:blank')
        return browser_arguments

    def _compute_launch_args(
        self,
        headless: bool = None,
        user_data_dir: str = None,
        devtools: bool = False,
        ignore_default_args: bool | list[str] = False,
        proxy: str = None,
        args: list[str] = None,
    ) -> None:
        if not ignore_default_args:
            self.browser_arguments.extend(
                self._default_args(headless, args, user_data_dir, devtools)
            )
        elif isinstance(ignore_default_args, list):
            self.browser_arguments.extend(
                filter(
                    lambda arg: arg not in ignore_default_args,
                    self._default_args(headless, args, user_data_dir, devtools)
                )
            )
        else:
            self.browser_arguments.extend(args)

    @staticmethod
    def _print_needs_install_message(binary: str) -> None:
        install_error = (
            f"Failed to find browser binary at {binary}\n"
            f'\n\t{"*" * 53}'
            "\n\t* Please install binary via \"mokr install\" or       *"
            "\n\t* specify the target binary path via \"binary_path\"  *"
            "\n\t* in the \"launch\" command.                          *"
            "\n\t*                                       >>>✈        *"
            f'\n\t{"*" * 53}'
        )
        raise FileNotFoundError(install_error)

    def _parse_proxy(self, proxy: str) -> None:
        raise NotImplementedError

    def _get_free_port(self) -> int:
        sock = socket.socket()
        sock.bind(('localhost', 0))
        port = sock.getsockname()[1]
        sock.close()
        del sock
        gc.collect()
        return port

    async def stop(self) -> None:
        """Stop the browser process if it is running."""
        if not self.browser_closed:
            await self.kill_browser()

    def _initial_page_callback(self) -> None:
        self.initial_page_promise.set_result(True)

    def _check_target(self, target: Target) -> None:
        if target.kind == 'page':
            self._initial_page_callback()

    async def launch(self) -> Browser:
        """Start browser process and return a `mokr.browser.Browser` object."""
        self.browser_closed = False
        self.connection = None
        options = {}
        options['env'] = self.env
        if not self.dumpio:
            options['stdout'] = subprocess.DEVNULL
            options['stderr'] = subprocess.STDOUT
        self.proc = subprocess.Popen(self.cmd, **options)
        self.browser_ws_endpoint = get_ws_endpoint(self.url)
        LOGGER.info(f'Browser listening on: {self.browser_ws_endpoint}')
        self.connection = Connection(
            self.browser_ws_endpoint,
            self._loop,
            self.slow_mo,
        )
        browser = Browser(
            self.kind,
            self.connection,
            [],
            self.ignore_https_errors,
            self.default_viewport,
            self.proc,
            self.kill_browser,
            self.proxy_credentials,
            self.default_user_agent,
        )
        await browser.ready()
        await self.ensure_initial_page(browser)
        return browser

    def _wait_for_browser_close(self) -> None:
        if self.proc.poll() is None and not self.browser_closed:
            self.browser_closed = True
            try:
                self.proc.terminate()
                self.proc.wait()
            except Exception:
                # Browser process may be already closed.
                pass

    async def ensure_initial_page(self, browser: Browser) -> None:
        """
        Wait for a new page in a given `browser` to be created.

        Args:
            browser (Browser): Target `mokr.browser.Browser`.
        """
        for target in browser.targets():
            if target.kind == 'page':
                return
        listeners = [
            add_event_listener(browser, 'targetcreated', self._check_target)
        ]
        await self.initial_page_promise
        remove_event_listeners(listeners)

    async def kill_browser(self) -> None:
        """Kill running browser process."""
        LOGGER.info('Killing browser process...')
        if self.connection and self.connection._connected:
            try:
                await self.connection.send(BROWSER_CLOSE)
                await self.connection.dispose()
            except Exception:
                LOGGER.warning("Ignored error killing browser.", exc_info=True)
        self._wait_for_browser_close()
        self._clean_restore_data_dirs()



from __future__ import annotations

import asyncio
import base64
import json
import logging
import re
from typing import TYPE_CHECKING, Awaitable

import httpx

from mokr.connection import DevtoolsConnection
from mokr.constants import (
    HTTP_RESPONSE_CODES_TEXT_MAP,
    NETWORK_GET_RESPONSE_BODY,
)
from mokr.exceptions import FirefoxNotImplementedError
from mokr.network.security import SecurityDetails

if TYPE_CHECKING:
    from mokr.network.request import Request


LOGGER = logging.getLogger(__name__)


class Response():
    def __init__(
        self,
        client: DevtoolsConnection,
        request: Request,
        status: int,
        headers: dict[str, str],
        from_disk_cache: bool,
        from_service_worker: bool,
        from_firefox: bool,
        security_details: dict = None,
        extra_info: dict = None,
        httpx_response: httpx.Response | None = None,
    ) -> None:
        """
        Reprentation of a remote response to a network request. `Response`
        objects are created by the `mokr.network.NetworkManager` when a response
        or redirect response is received.

        Pseudo `Response` objects are also made from `httpx.Response` objects
        by a `mokr.network.HttpDomain`.

        Args:
            client (DevtoolsConnection): A `mokr.connection.DevtoolsConnection`
                spawned by the parent `mokr.network.NetworkManager`.
            request (Request): The `mokr.request.Request` that this `Response`
                is bound to. Importantly, the `Request` does not spawn this.
            status (int): Status code for the response.
            headers (dict[str, str]): Response headers as dictionary.
            from_disk_cache (bool): Whether the remote response was served
                from the cache or not.
            from_service_worker (bool): Whether the remote response was served
                by a service worker or not.
            security_details (dict, optional): Security information for a
                given response. Defaults to None.
            extra_info (dict, optional): Any extra info associated with the
                remote response. Defaults to None.
            httpx_response: (httpx.Response | None, optional): If created by a
                `mokr.network.HttpDomain`, the original `httpx.Response` that
                this will be built from.
        """
        self._client = client
        self._request = request
        self._content_promise = self._client._loop.create_future()
        self._url = request.url
        self._from_disk_cache = from_disk_cache
        self._from_service_worker = from_service_worker
        self._from_firefox = from_firefox
        if not extra_info:
            extra_info = {}
        self._status = status
        self._reason = HTTP_RESPONSE_CODES_TEXT_MAP.get(str(status), "Unknown")
        self._extra_status_info = self._parse_extra_info_status_text(extra_info)
        _headers = extra_info.get("headers", headers)
        self._headers = {k.lower(): v for k, v in _headers.items()}
        self._security_details: dict | SecurityDetails = {}
        if security_details:
            self._security_details = SecurityDetails(
                security_details['subjectName'],
                security_details['issuer'],
                security_details['validFrom'],
                security_details['validTo'],
                security_details['protocol'],
            )
        self._httpx_response = httpx_response
        # Overridden via Firefox fetch.
        self._content = None

    def __repr__(self) -> str:
        return f"<Response ({self.status}: {self.reason}) at {id(self)}>"

    @staticmethod
    def _parse_extra_info_status_text(extra_info: dict = None) -> str | None:
        if not extra_info or not extra_info.get("headersText"):
            return
        first_line = extra_info["headersText"].split('\r')[0]
        if not first_line:
            return
        match = re.match(r"/[^ ]* [^ ]* (.*)/", first_line)
        if not match:
            return
        status_text = match[1]
        if not status_text:
            return
        return status_text

    @property
    def url(self) -> str:
        """URL from the remote response."""
        return self._url

    @property
    def ok(self) -> bool:
        """
        Whether the response failed or not (status code between 399 and 600).
        """
        return self._status == 0 or not 400 <= self._status < 599

    @property
    def status(self) -> int:
        """The HTTP status code for the remote response."""
        return self._status

    @property
    def extra_status_info(self) -> str | None:
        """Additional status information from an event's "extraInfo" or None."""
        return self._extra_status_info

    @property
    def reason(self) -> str:
        """The HTTP reason text for the remote response."""
        return self._reason

    @property
    def headers(self) -> dict:
        """All headers for the target response with keys lowered."""
        return self._headers

    @property
    def security_details(self) -> SecurityDetails | None:
        """
        Initialised `mokr.network.SecurityDetails` if the remote response
        was securely received, otherwise None.
        """
        return self._security_details

    @property
    def request(self) -> Request:
        """The `mokr.network.Request` object bound to this `Response`."""
        return self._request

    @property
    def from_cache(self) -> bool:
        """
        Whether the remote response was served from the disk or memory caches.
        """
        return self._from_disk_cache or self._request._from_memory_cache

    @property
    def from_service_worker(self) -> bool:
        """
        Whether the remote response was served by a service worker or not.
        """
        return self._from_service_worker

    @property
    def httpx_response(self) -> bool:
        """
        If created by a `mokr.network.HttpDomain`, this will hold the original
        `httpx.Response` object this was populated from. Otherwise, None.
        """
        return self._httpx_response

    async def _bufread(self) -> bytes:
        response = None
        try:
            response = await self._client.send(
                NETWORK_GET_RESPONSE_BODY,
                {'requestId': self._request._request_id},
            )
        except Exception as e:
            # Redirects have no body.
            if "No data found for resource with given identifier" in str(e):
                pass
            elif 'No resource with given identifier found' in str(e):
                raise type(e)(
                    'Could not load response body, request may be preflight.'
                )
            else:
                raise e
        if not response:
            body = b''
        else:
            body = response.get('body', b'')
            if response.get('base64Encoded'):
                return base64.b64decode(body)
        return body

    def buffer(self, force: bool = False) -> Awaitable[bytes]:
        """
        Return buffer awaitable which queries remote response object's body, if
        not already loaded in this `Response`. Can force querying with the
        `force` option.

        Args:
            force (bool, optional): Whether to bypass checking if the body
                has been cached here. Useful for `mokr.network.FetchDomain`
                requests or other requests where the body may resolve slower.
                Defaults to False.

        Returns:
            Awaitable[bytes]: Awaitable that yields response body as bytes.
        """
        if self._content:
            return self._content
        elif self._from_firefox:
            raise FirefoxNotImplementedError(
                "Firefox doesn't support functionality to use method"
                " Response.buffer."
            )
        elif self._httpx_response:
            # See note in HttpDomain init.
            return asyncio.to_thread(self._httpx_response.read)
        elif force or not self._content_promise.done():
            self._content_promise = self._client._loop.create_task(
                self._bufread()
            )
        return self._content_promise

    async def content(self) -> str | bytes:
        """
        Get body of the remote response object.

        If content is empty, tries to rerun query for the remote response
        object's body via `Response.buffer(force=True)`.

        Returns:
            str | bytes: Content of response.
        """
        if self._content:
            return self._content
        content = await self.buffer()
        if not content:
            # Content may not be accessible yet, retry.
            content = await self.buffer(force=True)
        if isinstance(content, str):
            return content
        else:
            try:
                return content.decode('utf-8')
            except UnicodeDecodeError:
                return content

    async def json(self) -> dict:
        """
        Load the body via `json.loads`.

        Raises:
            JSONDecodeError: Raised from `json.loads` if body is not JSON text.

        Returns:
            dict: Loaded body as dictionary.
        """
        content = await self.content()
        return json.loads(content)

    async def to_dict(self) -> dict:
        """
        Return a dictionary representation of the `Response` including the
        `Response.status`, `Response.headers`, and the response body via
        `Response.content`.

        Returns:
            dict: _description_
        """
        return {
            "status": self.status,
            "headers": self.headers,
            "body": await self.content(),
        }



import asyncio
import concurrent.futures
from typing import Any, Awaitable

from mokr.constants import (
    FRAME_DETACHED,
    FRAME_NAVIGATED_IN_DOC,
    LIFECYCLE_EVENT,
    LIFECYCLE_EVENTS,
)
from mokr.exceptions import MokrTimeoutError
from mokr.frame import Frame, FrameManager
from mokr.utils.remote import add_event_listener, remove_event_listeners


class NavigationWaiter:
    def __init__(
        self,
        frame_manager: FrameManager,
        frame: Frame,
        timeout: int,
        wait_until: list[LIFECYCLE_EVENTS] | LIFECYCLE_EVENTS,
    ) -> None:
        """
        Class to listen to a `mokr.frame.FrameManager` for specific navigation
        event(s) (`LIFECYCLE_EVENTS`) in the target `mokr.frame.Frame` for up
        to a given `timeout`. If all events not seen before timeout, raises
        a `MokrTimeoutError` exception.

        Args:
            frame_manager (FrameManager): The `mokr.frame.FrameManager` to watch
                for lifecycle events.
            frame (Frame): The `mokr.frame.Frame` that seen events count toward.
            timeout (int): Tiem in milliseconds to wait for event.
            wait_until (list[LIFECYCLE_EVENTS] | LIFECYCLE_EVENTS): Either a
                list of events or a single event to wait for.
        """
        if isinstance(wait_until, str):
            wait_until = [wait_until]
        self._validate_wait_until(wait_until)
        self._frame_manager = frame_manager
        self._frame = frame
        self._initial_loader_id = frame._loader_id
        self._timeout = timeout
        self._timeout_timer: asyncio.Task | asyncio.Future | None = None
        self._has_same_document_navigation = False
        self._event_listeners = [
            add_event_listener(
                self._frame_manager,
                LIFECYCLE_EVENT,
                self._check_lifecycle_complete,
            ),
            add_event_listener(
                self._frame_manager,
                FRAME_NAVIGATED_IN_DOC,
                self._navigated_within_document,
            ),
            add_event_listener(
                self._frame_manager,
                FRAME_DETACHED,
                self._check_lifecycle_complete,
            ),
        ]
        self._loop = self._frame_manager._client._loop
        self._lifecycle_complete_promise = self._loop.create_future()
        self._navigation_promise = self._loop.create_task(
            asyncio.wait(
                [
                    self._lifecycle_complete_promise,
                    self._create_timeout_promise(),
                ],
                return_when=concurrent.futures.FIRST_COMPLETED,
            )
        )
        self._navigation_promise.add_done_callback(lambda _: self._cleanup())

    def _validate_wait_until(self, wait_until: list[LIFECYCLE_EVENTS]) -> None:
        arg_name_to_lifecycle_event = {
            'load': 'load',
            'domcontentloaded': 'DOMContentLoaded',
            'documentloaded': 'DOMContentLoaded',
            'networkidle': 'networkIdle',
            'networkalmostidle': 'networkAlmostIdle',
        }
        self._expected_lifecycle = []
        for name in wait_until:
            protocol_event = arg_name_to_lifecycle_event.get(name)
            if not protocol_event:
                raise ValueError(
                    f'Unknown value(s) given for wait_until: {name}'
                )
            self._expected_lifecycle.append(protocol_event)

    async def _timeout_func(self) -> None:
        error_message = f'Navigation timeout of {self._timeout}ms exceeded.'
        await asyncio.sleep(self._timeout / 1000)
        self._maximum_timer.set_exception(MokrTimeoutError(error_message))

    def _create_timeout_promise(self) -> Awaitable[None]:
        self._maximum_timer = self._loop.create_future()
        if self._timeout:
            self._timeout_timer = self._loop.create_task(self._timeout_func())
        else:
            self._timeout_timer = self._loop.create_future()
        return self._maximum_timer

    def _navigated_within_document(self, frame: Frame = None) -> None:
        if frame != self._frame:
            return
        self._has_same_document_navigation = True
        self._check_lifecycle_complete()

    def _check_lifecycle_complete(self, *args) -> None:
        if (
            self._frame._loader_id == self._initial_loader_id
            and not self._has_same_document_navigation
        ):
            return
        elif not self._check_lifecycle(self._frame, self._expected_lifecycle):
            return
        elif not self._lifecycle_complete_promise.done():
            self._lifecycle_complete_promise.set_result(None)

    def _check_lifecycle(
        self,
        frame: Frame,
        expected_lifecycle: list[str],
    ) -> bool:
        if any(
            event not in frame._lifecycle_events for event in expected_lifecycle
        ):
            return False
        for child in frame.child_frames:
            if not self._check_lifecycle(child, expected_lifecycle):
                return False
        return True

    def _cleanup(self) -> None:
        remove_event_listeners(self._event_listeners)
        self._lifecycle_complete_promise.cancel()
        self._maximum_timer.cancel()
        self._timeout_timer.cancel()

    def navigation_promise(self) -> Any:
        """
        Return the promise so errors can be handled externally.

        Returns:
            Any: Navigation promise.
        """
        return self._navigation_promise

    def cancel(self) -> None:
        """Stop this waiter, raise no errors."""
        self._cleanup()
